import torch
from typing import Callable

from gpytorch.models import ApproximateGP
from gpytorch.distributions import MultivariateNormal, MultitaskMultivariateNormal
from torch_fenics import FEniCSModule

from alfi.models import VariationalLFM, TrainMode
from alfi.configuration import VariationalConfiguration
from alfi.utilities.torch import softplus


class PartialLFM(VariationalLFM):
    def __init__(self,
                 num_outputs,
                 gp_model: ApproximateGP,
                 fenics_model: FEniCSModule,
                 fenics_parameters: list,
                 config: VariationalConfiguration,
                 num_training_points=None,
                 dtype=torch.float64):
        super().__init__(num_outputs, gp_model, config, num_training_points, dtype)
        if self.config.initial_conditions:
            raise Exception('Initial conditions are not implemented for PartialLFM.')
        self.fenics_model = fenics_model
        self.fenics_parameters = fenics_parameters
        self.fenics_named_parameters = dict()
        name = 0
        for parameter in self.fenics_parameters:
            self.register_parameter(name='fenics' + str(name), param=parameter)
            self.fenics_named_parameters['fenics' + str(name)] = parameter
            name += 1

    def forward(self, tx, step_size=1e-1, return_samples=False, step=1, **kwargs):
        """
        tx : torch.Tensor
            Shape (2, num_times) or, if in pretrain mode, then a tuple containing input and output
        h : torch.Tensor the initial state of the ODE
            Shape (num_genes, 1)
        Returns
        -------
        Returns evolved h across times t.
        Shape (num_genes, num_points).
        """
        self.nfe = 0
        # self.gp_model.share_memory()
        # Get GP outputs
        if self.train_mode == TrainMode.GRADIENT_MATCH:
            t_f = tx[0].transpose(0, 1)
            data = tx[0]
        else:
            t_f = tx.transpose(0, 1)
            data = tx
        num_t = data[0, :].unique().shape[0]
        num_x = data[1, :].unique().shape[0]

        q_u = self.gp_model(t_f)
        u = q_u.rsample(torch.Size([self.config.num_samples])).permute(0, 2, 1)
        u = self.G(u)  # (S, num_outputs, tx)
        u = u.view(*u.shape[:2], num_t, num_x)
        if self.train_mode == TrainMode.GRADIENT_MATCH:
            params = [softplus(param.repeat(self.config.num_samples, 1)) for param in self.fenics_parameters]
            outputs = kwargs['pde_func'](tx[1], u[:, :, ::step].contiguous(), *params)
        else:
            outputs = self.solve_pde(u, step=step)

        if return_samples:
            return outputs

        f_mean = outputs.mean(dim=0).view(1, -1)  # shape (batch, times * distance)
        f_var = outputs.var(dim=0).view(1, -1) + 1e-7
        # TODO: make distribution something less constraining
        f_covar = torch.diag_embed(f_var)
        batch_mvn = MultivariateNormal(f_mean, f_covar)
        return MultitaskMultivariateNormal.from_batch_mvn(batch_mvn, task_dim=0)

    def func(self, i, u, step):
        # i, u = iu
        fenics_model = self.fenics_model
        time_steps = fenics_model.time_steps
        mesh_cells = fenics_model.mesh.cells().shape[0]

        # Integrate forward from the initial positions h0.
        outputs = list()
        y_prev = torch.zeros((1, mesh_cells + 1), requires_grad=False, dtype=torch.float64)
        params = [softplus(param) for param in self.fenics_parameters]

        # t = df['t'].values[:41]
        for n in range((time_steps + 1)):
            u_n = u[i, 0, n].unsqueeze(0)  # (S, t)
            # print(u_n.shape, y_prev.shape, params[0].shape)
            y_prev = fenics_model(y_prev, u_n, *params)

            # y_prev shape (N, 21)
            outputs.append(y_prev)

        outputs = torch.stack(outputs).permute(1, 0, 2)  # (S, T, X)
        outputs = outputs[:, ::step, :]
        return outputs

    def solve_pde(self, u, step=1):
        """

        @param u: Shape (S, 1, num_t, num_x)
        @return:
        """
        # u.share_memory_()
        outputs = [self.func(i, u, step) for i in range(self.config.num_samples)]

        outputs = torch.cat(outputs)
        # outputs = torch.cat(self.pool.map(self.func, [() for i in range(self.config.num_samples)]))  #self.pool

        return outputs

    def G(self, u):
        return u



class Parameter:
    def __init__(self,
                 name,
                 prior,
                 initial_value,
                 transform=None,
                 fixed=False):
        self.name = name
        self.prior = prior
        self.transform = (lambda x: x) if transform is None else transform
        self.value = initial_value
        self.fixed = fixed

    def propose(self, *args):
        if self.fixed:
            return self.value
        assert self.proposal_dist is not None, 'proposal_dist must not be None if you use propose()'
        return self.proposal_dist(*args).sample().numpy()



from .lfm import LFM
from .exact_lfm import ExactLFM
from .variational_lfm import VariationalLFM, TrainMode
from .ordinary_lfm import OrdinaryLFM
from .ordinary_lfm_no_precompute import OrdinaryLFMNoPrecompute
from .approximate_gp import MultiOutputGP, generate_multioutput_gp
from .operator import NeuralOperator
from .recurrent_operator import RecurrentNeuralOperator
from .neural_lfm import NeuralLFM


modules = [
    'LFM',
    'NeuralLFM',
    'ExactLFM',
    'VariationalLFM',
    'OrdinaryLFM',
    'OrdinaryLFMNoPrecompute',
    'NeuralOperator',
    'RecurrentNeuralOperator',
    'MultiOutputGP',
    'generate_multioutput_gp',
    'TrainMode',
]

try:
    import fenics
    import fenics_adjoint
    import torch_fenics
    fenics_present = True
except ImportError as e:
    print(e)
    fenics_present = False


if fenics_present:
    from .partial_lfm import PartialLFM
    modules.extend([
        'PartialLFM',
    ])


__all__ = modules



import torch
from torch.nn import Parameter
from gpytorch.constraints import Positive

from alfi.configuration import VariationalConfiguration
from alfi.models import OrdinaryLFM, TrainMode


class TranscriptionLFM(OrdinaryLFM):
    def __init__(self, num_outputs, gp_model, config: VariationalConfiguration,
                 initial_basal=None, initial_decay=None, initial_sensitivity=None, **kwargs):
        super().__init__(num_outputs, gp_model, config, **kwargs)
        self.positivity = Positive()
        initial_basal = 0.1 if initial_basal is None else initial_basal
        initial_decay = 0.3 if initial_decay is None else initial_decay
        initial_sensitivity = 1 if initial_sensitivity is None else initial_sensitivity
        self.raw_decay = Parameter(self.positivity.inverse_transform(
            initial_decay + torch.rand(torch.Size([self.num_outputs, 1]), dtype=torch.float64)))
        self.raw_basal = Parameter(self.positivity.inverse_transform(
            initial_basal * torch.rand(torch.Size([self.num_outputs, 1]), dtype=torch.float64)))
        self.raw_sensitivity = Parameter(self.positivity.inverse_transform(
            initial_sensitivity * torch.rand(torch.Size([self.num_outputs, 1]), dtype=torch.float64)))

    @property
    def decay_rate(self):
        return self.positivity.transform(self.raw_decay)

    @decay_rate.setter
    def decay_rate(self, value):
        self.raw_decay = self.positivity.inverse_transform(value)

    @property
    def basal_rate(self):
        return self.positivity.transform(self.raw_basal)

    @basal_rate.setter
    def basal_rate(self, value):
        self.raw_basal = self.positivity.inverse_transform(value)

    @property
    def sensitivity(self):
        return self.positivity.transform(self.raw_sensitivity)

    @sensitivity.setter
    def sensitivity(self, value):
        self.raw_sensitivity = self.decay_constraint.inverse_transform(value)

    def initial_state(self):
        return self.basal_rate / self.decay_rate

    def odefunc(self, t, h):
        """h is of shape (num_samples, num_outputs, 1)"""
        self.nfe += 1

        f = self.latent_gp
        if not (self.train_mode == TrainMode.GRADIENT_MATCH):
            f = self.latent_gp[:, :, self.t_index].unsqueeze(2)
            if t > self.last_t:
                self.t_index += 1
            self.last_t = t

        dh = self.basal_rate + self.sensitivity * f - self.decay_rate * h
        return dh



import torch
import numpy as np


def flatten_dataset(dataset):
    num_genes = dataset.num_outputs
    train_t = dataset[0][0]
    num_times = train_t.shape[0]
    m_observed = torch.stack([
        dataset[i][1] for i in range(num_genes)
    ]).view(num_genes, num_times)
    train_t = train_t.repeat(num_genes)
    train_y = m_observed.view(-1)
    return train_t, train_y



def dros_ground_truth(gene):
    """Becker et al."""
    kni_params = dict(sensitivity=0.0783,
                      decay=0.0770,
                      diffusion=0.0125)
    gt_params = dict(sensitivity=0.1107,
                     decay=0.1110,
                     diffusion=0.0159)
    kr_params = dict(sensitivity=0.0970,
                     decay=0.0764,
                     diffusion=0.0015)
    return dict(kr=kr_params, kni=kni_params, gt=gt_params)[gene]

def hafner_ground_truth():
    s = np.array([
        0.232461671, 0.429175332, 1.913169606, 0.569821512, 2.139812962, 0.340465324,
        4.203117214, 0.635328943, 0.920901229, 0.263968666, 1.360004451, 4.816673998,
        0.294392325, 2.281036308, 0.86918333, 2.025737447, 1.225920534, 11.39455009,
        4.229758095, 4.002484948, 32.89511304, 7.836815916])
    d = np.array([
        0.260354271, 0.253728801, 0.268641114, 0.153037374, 0.472215028, 0.185626363,
        0.210251586, 0.211915623, 0.324826082, 0.207834775, 0.322725728, 0.370265667,
        0.221598164, 0.226897275, 0.409710437, 0.398004589, 0.357308033, 0.498836353,
        0.592101838, 0.284200056, 0.399638904, 0.463468107])
    b = np.zeros_like(d)
    return b, s, d


def create_tx(linspace, num_tasks=1, dtype=torch.float64):
    """
    Creates a 2D spatiotemporal input space.
    :param linspace: the linspace for both space and time
    :param num_tasks: the size of the batch dimension
    :return:
    """
    t = torch.repeat_interleave(linspace, linspace.shape[0])
    x = torch.tile(linspace, (linspace.shape[0], ))
    tx = torch.stack([t, x], dim=1)
    tx = tx.view(1, -1, 2).repeat(num_tasks, 1, 1)
    return tx.type(dtype)


def generate_neural_dataset_2d(txf, params, ntrain, ntest, sub=1):
    tx = txf[0, 0:2]
    s1 = tx[0, :].unique().shape[0]
    s2 = tx[1, :].unique().shape[0]
    grid = tx.t()
    grid = grid.reshape(1, s1, s2, 2).type(torch.float)

    data = txf.permute(0, 2, 1)
    data = data.reshape(data.shape[0], s1, s2, 4).type(torch.float)
    grid = grid[:, ::sub, ::sub, :]
    data = data[:, ::sub, ::sub, :]
    s1 = data.shape[1]
    s2 = data.shape[2]
    y_train = data[:ntrain, ..., 2:3]
    x_train = data[:ntrain, ..., 3]
    y_test = data[ntrain:, ..., 2:3]
    x_test = data[ntrain:, ..., 3]

    x_train = torch.cat([x_train.reshape(ntrain, s1, s2, 1), grid.repeat(ntrain, 1, 1, 1)], dim=3)
    x_test = torch.cat([x_test.reshape(ntest, s1, s2, 1), grid.repeat(ntest, 1, 1, 1)], dim=3)

    train = [(x_train[i], y_train[i], params[i].type(torch.float)) for i in range(ntrain)]
    test = [(x_test[i], y_test[i], params[i].type(torch.float)) for i in range(ntest)]

    return train, test


def generate_neural_dataset_1d(t_observed, datasets, params, ntrain=1, ntest=0):
    x_train = torch.cat([obs[0] for obs in datasets[:ntrain]]).permute(0, 2, 1).type(torch.float32)
    y_train = torch.cat([obs[1] for obs in datasets[:ntrain]]).permute(0, 2, 1).type(torch.float32)

    x_test = x_train
    y_test = x_test
    if ntest > 0:
        x_test = torch.cat([obs[0] for obs in datasets[ntrain:]]).permute(0, 2, 1).type(torch.float32)
        y_test = torch.cat([obs[1] for obs in datasets[ntrain:]]).permute(0, 2, 1).type(torch.float32)

    grid = t_observed.reshape(1, -1, 1).repeat(ntrain, 1, 1)  # (1, 32, 32, 40, 1)
    grid_test = t_observed.reshape(1, -1, 1).repeat(max(ntest, 1), 1, 1)  # (1, 32, 32, 40, 1)

    params = torch.stack(params, dim=0)
    params_train = params[:ntrain].squeeze(-1)
    params_test = params[ntrain:].squeeze(-1)
    x_train = torch.cat([grid, x_train], dim=-1)
    x_test = torch.cat([grid_test, x_test], dim=-1)

    # return x_train, x_test, y_train, y_test, params_train, params_test
    train = [(x_train[i], y_train[i], params[i].type(torch.float)) for i in range(ntrain)]
    test = [(x_test[i], y_test[i], params[i].type(torch.float)) for i in range(ntest)]

    return train, test


def context_target_split(x, y, num_context, num_extra_target, locations=None):
    """
    Given inputs x and their value y, return random subsets of points for
    context and target. Note that following conventions from "Empirical
    Evaluation of Neural Process Objectives" the context points are chosen as a
    subset of the target points.

    Parameters
    ----------
    x : torch.Tensor
        Shape (batch_size, num_points, x_dim)

    y : torch.Tensor
        Shape (batch_size, num_points, y_dim)

    num_context : int
        Number of context points.

    num_extra_target : int
        Number of additional target points.
    """
    num_points = x.shape[1]

    # Sample locations of context and target points
    if locations is None:
        locations = np.random.choice(num_points,
                                     size=num_context + num_extra_target, replace=False)

    if x.ndim > 3:
        x_context = x[:, locations[:num_context], ...]
        y_context = y[:, locations[:num_context], ...]
        x_target = x[:, locations, ...]
        y_target = y[:, locations, ...]
    else:
        x_context = x[:, locations[:num_context], :]
        y_context = y[:, locations[:num_context], :]
        x_target = x[:, locations, :]
        y_target = y[:, locations, :]
    return x_context, y_context, x_target, y_target



import tensorflow as tf
import torch
import pickle as pkl

from sklearn import preprocessing
from scipy.interpolate import interp1d
from tensorflow import math as tfm

from alfi.datasets import DataHolder
from alfi.mcmc import TranscriptionLikelihood
from alfi.configuration import MCMCConfiguration

from alfi.mcmc.models import TranscriptionMixedSampler
from alfi.utilities.tf import discretise, logistic, inverse_positivity
from matplotlib import pyplot as plt

import numpy as np

f64 = np.float64

#%%

def artificial_dataset(opt, likelihood_class, t_end=12, num_genes=13, num_tfs=3, weights=None, delays=[1, 4, 10], true_kbar=None):
    t = np.arange(t_end)
    τ, common_indices = discretise(t)
    time = (t, τ, tf.constant(common_indices))
    N_p = τ.shape[0]
    N_m = t.shape[0]

    # Transcription factors
    C = np.array([0.01, 0.2, 0.47, 0.51, 0.35, 0.19, 0.17, 0.24, 0.37, 0.47, 0.39, 0.2, 0.05, 0.025, 0.025, 0.01, 0.005, 0.005])
    A = np.array([0.01, 0.1, 0.22, 0.44, 0.53, 0.41, 0.23, 0.13, 0.09, 0.035, 0.022, 0.02, 0.015, 0.01, 0.005, 0.005])
    B = np.array([0.01, 0.02, 0.03, 0.05, 0.08, 0.16, 0.4, 0.36, 0.23, 0.12, 0.05, 0.025, 0.025, 0.01, 0.005, 0.005])
    interp = interp1d(np.arange(A.shape[0]), A, kind='cubic')
    A = interp(np.linspace(0,14, τ.shape[0]))
    interp = interp1d(np.arange(B.shape[0]), B, kind='cubic')
    B = interp(np.linspace(0,14, τ.shape[0]))
    interp = interp1d(np.arange(C.shape[0]), C, kind='quadratic')
    C = interp(np.linspace(0,14, τ.shape[0]))

    fbar = np.array([A, B, C])
    fbar = 15*preprocessing.normalize(fbar)
    fbar = np.expand_dims(fbar, 0)
    f_observed = tf.stack([fbar[:, i, common_indices] for i in range(num_tfs)], axis=1)
    fbar = tfm.log((tfm.exp(fbar)-1))
    f_i = tfm.log(1+tfm.exp(fbar))
    f_observed += tf.random.normal([N_m], stddev=tf.sqrt(0.02*f_observed), dtype='float64')
    # Kinetics
    true_k_fbar = logistic(f64(np.array([2, 2, 2]).T)) #a was [0.1, 0.1, 0.1]
    if true_kbar is None:
        true_kbar = (np.array([[0.50563, 0.66, 0.893, 0.9273],
                            [0.6402, 0.6335, 0.7390, 0.7714],
                            [0.6202, 0.6935, 0.7990, 0.7114],
                            [0.5328, 0.5603, 0.6498, 0.9244],
                            [0.5328, 0.6603, 0.6798, 0.8244],
                            [0.5939, 0.5821, 0.77716, 0.8387],
                            [0.50, 0.68, 0.75716, 0.8587],
                            [0.58, 0.67, 0.57, 0.95],
                            [0.5553, 0.5734, 0.6462, 0.9068],
                            [0.5750, 0.5548, 0.6380, 0.7347],
                            [0.5373, 0.5277, 0.6319, 0.8608],
                            [0.5372, 0.5131, 0.8000, 0.9004],
                            [0.5145, 0.5818, 0.6801, 0.9129]]))
    if weights is None:
        w = 1*tf.ones((num_genes, num_tfs), dtype='float64')
        w_0 = tf.zeros(num_genes, dtype='float64')
    else:
        w = weights[0]
        w_0 = weights[1]


    # Genes
    temp_data = DataHolder((np.ones((1, num_genes, N_m)), np.ones((1, num_tfs, N_m))), None, time)
    temp_lik = likelihood_class(temp_data, opt)
    
    Δ = tf.constant(delays, dtype='float64')
    m_pred = temp_lik.predict_m(true_kbar, true_k_fbar, (w), fbar, (w_0), Δ)

    m_observed = tf.stack([m_pred.numpy()[:,i,common_indices] for i in range(num_genes)], axis=1)
    m_observed += tf.random.normal([N_m], stddev=tf.sqrt(0.02*m_observed), dtype='float64')
    data = (m_observed, f_observed)

    data = DataHolder(data, None, time)

    return data, fbar, (true_kbar, true_k_fbar)


def get_artificial_dataset(num_genes=20, num_tfs=3):
    """
    Returns:
        nodelay, delay datasets
    """
    nodelay_path = '../data/articial_nodelay.pkl'
    delay_path = '../data/articial_delay.pkl'
    with open(nodelay_path, 'rb') as f:
        nodelay_dataset = pkl.load(f)
    with open(delay_path, 'rb') as f:
        delay_dataset = pkl.load(f)
    return nodelay_dataset, delay_dataset

    tf.random.set_seed(1)
    w = tf.random.normal([num_genes, num_tfs], mean=0.5, stddev=0.71, seed=42, dtype='float64')

    Δ_delay = tf.constant([0, 4, 10], dtype='float64')

    w_0 = tf.zeros(num_genes, dtype='float64')

    true_kbar = logistic((np.array([
        [1.319434062, 1.3962113525, 0.8245041865, 2.2684353378],
        [1.3080045137, 3.3992868747, 2.0189033658, 3.7460822389],
        [2.0189525448, 1.8480506624, 0.6805040228, 3.1039094120],
        [1.7758426875, 0.1907625023, 0.1925539427, 1.8306885751],
        [1.7207442227, 0.1252089546, 0.6297333943, 3.2567248923],
        [1.4878806850, 3.8623843570, 2.4816128746, 4.3931294404],
        [2.0853079514, 2.5115446790, 0.6560607356, 3.0945313562],
        [1.6144843688, 1.8651409657, 0.7785363895, 2.6845058360],
        [1.4858223122, 0.5396687493, 0.5842698019, 3.0026805243],
        [1.6610647522, 2.0486340884, 0.9863876546, 1.4300094581],
        [1.6027276189, 1.4320302060, 0.7175033248, 3.2151637970],
        [2.4912882714, 2.7935526605, 1.2438786874, 4.3944794204],
        [2.894114279, 1.4726280947, 0.7356719860, 2.2316019158],
     [1.7927833839, 1.0405867396, 0.4055775218, 2.9888350247],
     [1.0429721112, 0.1011544950, 0.7330443670, 3.1936843755],
     [1.2519286771, 2.0617880701, 1.0759649567, 3.9406060364],
     [1.4297185709, 1.3578824015, 0.6037986912, 2.6512418604],
     [1.9344878813, 1.4235867760, 0.8226320338, 4.2847217252],
     [1.4325562449, 1.1940752177, 1.0556928599, 4.1850449557],
     [0.8911103971, 1.3560009300, 0.5643954823, 3.4300182328],
     [1.0269654997, 1.0788097511, 0.5268448648, 4.4793299593],
     [0.8378220502, 1.8148234459, 1.0167440138, 4.4903387696]]
    )))
    true_kbar = true_kbar[:num_genes]

    opt = MCMCConfiguration(preprocessing_variance=False,
                            latent_data_present=True,
                            kinetic_exponential=True,
                            weights=True,
                            initial_step_sizes={'logistic': 1e-8, 'latents': 10},
                            delays=True)

    data, fbar, kinetics = artificial_dataset(opt, TranscriptionLikelihood, num_genes=num_genes,
                                              weights=(w, w_0), delays=Δ_delay.numpy(), t_end=10,
                                              true_kbar=true_kbar[:num_genes])
    true_kbar, true_k_fbar = kinetics
    f_i = inverse_positivity(fbar)
    t_observed, t_discretised, common_indices = data.t_observed, data.t_discretised, data.common_indices

    common_indices = common_indices.numpy()

    model = TranscriptionMixedSampler(data, opt)

    # Transcription factor
    plt.title('TFs')
    for i in range(num_tfs):
        plt.plot(t_discretised, f_i[0, i], label=f'TF {i}')
        plt.scatter(t_observed, data.f_obs[0, i], marker='x')
    plt.xticks(np.arange(0, 10))
    plt.legend()
    print(t_discretised.shape)


    lik = model.likelihood
    Δ_nodelay = tf.constant([0, 0, 0], dtype='float64')
    m_pred = lik.predict_m(true_kbar, true_k_fbar, (w), fbar, (w_0), Δ_delay)
    m_pred_nodelay = lik.predict_m(true_kbar, true_k_fbar, (w), fbar, (w_0), Δ_nodelay)
    m_observed_nodelay = tf.stack([m_pred_nodelay.numpy()[:,i,common_indices] for i in range(num_genes)], axis=1)
    m_observed = tf.stack([m_pred.numpy()[:,i,common_indices] for i in range(num_genes)], axis=1)

    p_nodelay = lik.calculate_protein(fbar, true_k_fbar, Δ_nodelay)
    p = lik.calculate_protein(fbar, true_k_fbar, Δ_delay)

    nodelay_dataset = (np.array(p_nodelay), np.array(m_observed_nodelay))
    delay_dataset = (np.array(p), np.array(m_observed))

    with open(nodelay_path, 'wb') as f:
        pkl.dump(nodelay_dataset, f)
    with open(delay_path, 'wb') as f:
        pkl.dump(delay_dataset, f)

    return nodelay_dataset, delay_dataset



import torch
import torch_fenics

from fenics import *
from fenics_adjoint import *

from alfi.utilities.torch import spline_interpolate_gradient


class ReactionDiffusion(torch_fenics.FEniCSModule):
    def __init__(self, t_range: tuple, time_steps, mesh):
        """
        Parameters:
            t_range: tuple(low, high)
            dt: scalar float
            mesh_cells: Number of cells in the spatial mesh
        """
        super().__init__()
        self.time_steps = time_steps
        self.dt = ((t_range[1] - t_range[0]) / self.time_steps).item()
        self.mesh = mesh
        self.V = FunctionSpace(self.mesh, 'P', 1)

        # Create trial and test functions
        y = TrialFunction(self.V)
        self.v = TestFunction(self.V)

    def interpolated_gradient(self, tx, y_matrix, disc=1, plot=False):
        num_t = tx[0, :].unique().shape[0]
        num_x = tx[1, :].unique().shape[0]
        num_t_orig = y_matrix.shape[-2]
        num_x_orig = y_matrix.shape[-1]
        dy_t = list()
        for i in range(num_x_orig):
            t = tx[0][::num_x][::disc]
            y = y_matrix[:, i].unsqueeze(-1)
            t_interpolate, y_interpolate, y_grad, _ = \
                spline_interpolate_gradient(t, y)
            dy_t.append(y_grad)
        dy_t = torch.stack(dy_t)

        d2y_x = list()
        dy_x = list()
        for i in range(num_t_orig):
            t = tx[1][::disc][:num_x]
            y = y_matrix[i].unsqueeze(-1)
            t_interpolate, y_interpolate, y_grad, y_grad_2 = \
                spline_interpolate_gradient(t, y)
            d2y_x.append(y_grad_2)
            dy_x.append(y_grad)

        d2y_x = torch.stack(d2y_x)
        dy_x = torch.stack(dy_x)[..., ::10, 0].reshape(1, -1)
        d2y_x = d2y_x[..., ::10, 0].reshape(1, -1)
        dy_t = dy_t[..., ::10, 0].t().reshape(1, -1)

        def pde_func(y, u, sensitivity, decay, diffusion):
            # y (1, 1681) u (25, 1, 41, 41) s (25, 1)
            dy_t = (sensitivity * u.view(u.shape[0], -1) -
                    decay * y.view(1, -1) +
                    diffusion * d2y_x)
            return dy_t

        if plot:
            from alfi.plot import plot_spatiotemporal_data
            ts = tx[0, :].unique().numpy()
            xs = tx[1, :].unique().numpy()
            extent = [ts[0], ts[-1], xs[0], xs[-1]]
            axes = plot_spatiotemporal_data(
                [
                    y_matrix.view(num_t_orig, num_x_orig).t(),
                    dy_t.reshape(num_t_orig, num_x_orig).t(),
                    d2y_x.view(num_t_orig, num_x_orig).t(),
                ],
                extent, titles=['y', 'dy_t', 'd2y_x']
            )

        return pde_func, dy_t.t()

    def solve(self, y_prev, u, sensitivity, decay, diffusion):
        # Construct bilinear form (Arity = 2 (for both Trial and Test function))
        y = TrialFunction(self.V)
        self.a = (1 + self.dt * decay) * y * self.v * dx + self.dt * diffusion * inner(grad(y), grad(self.v)) * dx

        # Construct linear form
        L = (y_prev + self.dt * sensitivity * u) * self.v * dx

        # Construct boundary condition
        bc = DirichletBC(self.V, Constant(0), 'on_boundary')

        # Solve the equation
        y = Function(self.V)
        solve(self.a == L, y, bc)

        return y

    def input_templates(self):
        # Declare templates for the inputs to Poisson.solve
        return Function(self.V), Function(self.V), \
               Constant(0), Constant(0), Constant(0)



from gpytorch.mlls import VariationalELBO


class MaskedVariationalELBO(VariationalELBO):

    def _log_likelihood_term(self, variational_dist_f, target, mask=None, weight=None, **kwargs):
        log_prob = self.likelihood.expected_log_prob(target, variational_dist_f, **kwargs)
        if weight is not None:
            log_prob *= weight
        if mask is not None:
            log_prob = log_prob[mask]
        return log_prob.sum(-1)



import torch
import gpytorch

from gpytorch.models import ApproximateGP
from gpytorch.variational import (
    NaturalVariationalDistribution,
    CholeskyVariationalDistribution,
    VariationalStrategy,
    IndependentMultitaskVariationalStrategy,
    LMCVariationalStrategy,
    TrilNaturalVariationalDistribution
)


class MultiOutputGP(ApproximateGP):
    def __init__(self,
                 mean_module,
                 covar_module,
                 inducing_points,
                 num_tasks,
                 num_latents=None,  # optional since if independent, num_latents == num_tasks
                 learn_inducing_locations=False,
                 natural=True,
                 use_tril=False,
                 independent=True):
        if independent:
            num_latents = num_tasks
        assert num_latents is not None
        # The variational dist batch shape means we learn a different variational dist for each latent
        if natural:
            Distribution = TrilNaturalVariationalDistribution if use_tril else NaturalVariationalDistribution
            variational_distribution = Distribution(
                inducing_points.size(-2), batch_shape=torch.Size([num_latents])
            )
        else:
            variational_distribution = CholeskyVariationalDistribution(
                inducing_points.size(-2), batch_shape=torch.Size([num_latents])
            )

        # Wrap the VariationalStrategy in a MultiTask to make output MultitaskMultivariateNormal
        # rather than a batch MVN
        if independent:
            Strategy = IndependentMultitaskVariationalStrategy
            strategy_kwargs = dict()
        else:
            Strategy = LMCVariationalStrategy
            strategy_kwargs = dict(num_latents=num_latents, latent_dim=-1)

        variational_strategy = Strategy(
            VariationalStrategy(
                self, inducing_points, variational_distribution, learn_inducing_locations=learn_inducing_locations
            ),
            num_tasks=num_tasks,
            **strategy_kwargs
        )
        super().__init__(variational_strategy)

        self.mean_module = mean_module
        self.covar_module = covar_module

    def get_inducing_points(self):
        return self.variational_strategy.base_variational_strategy.inducing_points

    def forward(self, t):
        # The forward function should be written as if we were dealing with each output
        # dimension in batch
        mean_x = self.mean_module(t)
        covar_x = self.covar_module(t)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


def generate_multioutput_gp(num_tasks, inducing_points,
                            kernel_class=gpytorch.kernels.RBFKernel,
                            kernel_kwargs=None,
                            ard_dims=None,
                            use_scale=False,
                            initial_lengthscale=None,
                            lengthscale_constraint=None,
                            zero_mean=True,
                            gp_kwargs=None,
                            num_latents=None):
    # Modules should be marked as batch so different set of hyperparameters are learnt.
    if num_latents is None:  # this is None is we are using an independent variational strategy,
        num_latents = num_tasks  # in which case the #tasks == #latent GPs.
    if gp_kwargs is None:
        gp_kwargs = {}
    if kernel_kwargs is None:
        kernel_kwargs = {}
    if zero_mean:
        mean_module = gpytorch.means.ZeroMean(batch_shape=torch.Size([num_latents]))
    else:
        mean_module = gpytorch.means.ConstantMean(batch_shape=torch.Size([num_latents]))
    covar_module = kernel_class(
        batch_shape=torch.Size([num_latents]),
        ard_num_dims=ard_dims,
        lengthscale_constraint=lengthscale_constraint,
        **kernel_kwargs
    )
    if use_scale:
        covar_module = gpytorch.kernels.ScaleKernel(
            covar_module,
            batch_shape=torch.Size([num_latents])
        )
    if initial_lengthscale is not None:
        if use_scale:
            covar_module.base_kernel.lengthscale = initial_lengthscale
        else:
            covar_module.lengthscale = initial_lengthscale
    return MultiOutputGP(mean_module, covar_module, inducing_points, num_tasks, num_latents=num_latents, **gp_kwargs)



import operator
import torch

from functools import reduce
from torch.distributions import Normal
from torch.nn.functional import softplus

from alfi.nn import SimpleBlock1d, SimpleBlock2d
from alfi.models import LFM
from alfi.models import NeuralOperator


class RecurrentNeuralOperator(LFM):
    def __init__(self, block_dim, in_channels, out_channels, modes, width, num_layers=4):
        super().__init__()
        self.params = False
        self.operator = NeuralOperator(block_dim, in_channels, out_channels, modes, width,
                                       num_layers=num_layers, params=False)

    def forward(self, x, step=1):
        """
        x shape (batch, space, times, channels)
        """
        batch_size, T, S, _ = x.shape
        # Last dimension is time
        pred = None
        for t in range(0, T, step):
            y_pred = self.operator(x[:, t, ...]).unsqueeze(-1)  # (batch, space, channels, 1)
            if t == 0:
                pred = y_pred
            else:
                pred = torch.cat((pred, y_pred), -1)

            # x = torch.cat((x[..., step:], y_pred), dim=-1)

        return pred.permute(0, 3, 1, 2)

    def count_params(self):
        c = 0
        for p in self.parameters():
            c += reduce(operator.mul, list(p.size()))
        return c

    def predict_f(self, tx_predict):
        with torch.no_grad():
            y_pred = self(tx_predict)
            p_y_mean = y_pred[..., 0]
            p_y_var = softplus(y_pred[..., 1])
            p_y_pred = Normal(p_y_mean, p_y_var)
            return p_y_pred

    def save(self, filepath):
        torch.save(self.state_dict(), str(filepath) + 'lfo.pt')



import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow_probability import distributions as tfd

from alfi.mcmc.results import GenericResults
from .mixins import ParamGroupMixin

class GibbsSampler(tfp.mcmc.TransitionKernel, ParamGroupMixin):
    
    def __init__(self, param, sq_diff_fn, N):
        super().__init__()
        self.param_group = [param]
        self.prior = param.prior
        self.sq_diff_fn = sq_diff_fn
        self.N = N

    @tf.function
    def one_step(self, current_state, previous_kernel_results):
        # Prior parameters
        α = self.prior.concentration
        β = self.prior.scale
        # Conditional posterior of inv gamma parameters:
        sq_diff = self.sq_diff_fn()
        α_post = α + 0.5*self.N
        β_post = β + 0.5*tf.reduce_sum(sq_diff, axis=1)
        # print(α.shape, sq_diff.shape)
        # print('val', β_post.shape, params.σ2_m.value)
        new_state = tfd.InverseGamma(α_post, β_post).sample()
        new_state = tf.reshape(new_state, (sq_diff.shape[0], 1))
        return [new_state], GenericResults(list(), True)

    def bootstrap_results(self, init_state):
        return GenericResults(list(), True) 
    
    def is_calibrated(self):
        return True




import seaborn as sns


palette = sns.color_palette('colorblind')


class Colours:
    shade_color = palette[0]
    line_color = palette[0]
    line2_color = palette[2]
    shade2_color = palette[2]
    scatter_color = palette[3]
    bar1_color = palette[3]
    bar2_color = palette[2]
    error_color = palette[1]



import tensorflow as tf
from tensorflow_probability import distributions as tfd

from alfi.mcmc.gp.gp_kernels import GPKernelSelector
from alfi.mcmc.samplers import MetropolisKernel
from alfi.utilities.tf import jitter_cholesky, logit, add_diag
from alfi.mcmc.results import GenericResults
from .mixins import ParamGroupMixin

import numpy as np
f64 = np.float64


class LatentGPSampler(MetropolisKernel, ParamGroupMixin):
    """
    Latent Gaussian Process sampler.
    Parameters:
        likelihood_fn:
        latent_likelihood_fn:
        param: the associated Parameter (only one, not a list)
    """
    def __init__(self, likelihood_fn, latent_likelihood_fn,
                 param,
                 kernel_selector: GPKernelSelector, 
                 step_size, kernel_exponential=False):
        self.param_group = [param]
        self.fbar_prior_params = kernel_selector()
        self.kernel_priors = kernel_selector.priors()
        self.kernel_selector = kernel_selector
        self.likelihood_fn = likelihood_fn
        self.latent_likelihood_fn = latent_likelihood_fn
        self.kernel_exponential = kernel_exponential
        self.data_present = latent_likelihood_fn is not None
        self.step_fn = self.joint_one_step
        self.calc_prob_fn = self.joint_calc_prob

        super().__init__(step_size, tune_every=3000)

    def _one_step(self, current_state, previous_kernel_results):
        return self.step_fn(current_state, previous_kernel_results)

    @tf.function
    def joint_one_step(self, current_state, previous_kernel_results):
        current_state = current_state[0]
        new_state = tf.identity(current_state[0])
        num_replicates = new_state.shape[0]
        num_tfs = new_state.shape[1]
        new_params = []
        S = tf.linalg.diag(self.step_size)
        # MH
        m, K = self.fbar_prior_params(current_state[1], current_state[2])
        N_p = K.shape[-1]
        K = K+tf.linalg.diag(1e-7*tf.ones(N_p, dtype='float64'))

        # Propose new params
        v = self.kernel_selector.proposal(0, current_state[1]).sample()
        l2 = self.kernel_selector.proposal(1, current_state[2]).sample()
        m_, K_ = self.fbar_prior_params(v, l2)
        K_ = K_+tf.linalg.diag(1e-7*tf.ones(N_p, dtype='float64'))
        iK = tf.linalg.inv(K)
        iK_ = tf.linalg.inv(K_)
        U_invR = tf.linalg.cholesky(add_diag(iK, 1/S))
        U_invR = tf.transpose(U_invR, [0, 2, 1])
        U_invR_ = jitter_cholesky(add_diag(iK_, 1/S))
        U_invR_ = tf.transpose(U_invR_, [0, 2, 1])
        new_hyp = [v, l2]
        old_hyp = [current_state[1], current_state[2]]

        # Gibbs step
        fbar = new_state
        for r in range(num_replicates):
            fbar = new_state[r]

            gg = tfd.MultivariateNormalDiag(fbar, self.step_size).sample()

            Sinv_g = gg / self.step_size

            nu = tf.linalg.matvec(U_invR, fbar) - tf.squeeze(tf.linalg.solve(tf.transpose(U_invR, [0, 2, 1]), tf.expand_dims(Sinv_g, -1)), -1)
            f = tf.linalg.solve(U_invR_, tf.expand_dims(nu, -1)) + tf.linalg.cholesky_solve(tf.transpose(U_invR_, [0, 2, 1]), tf.expand_dims(Sinv_g, -1))
            f = tf.squeeze(f, -1)

            mask = np.zeros((num_replicates, 1, 1), dtype='float64')
            mask[r] = 1
            new_state = (1-mask) * new_state + mask * f
        
        test_state = tf.zeros((num_replicates, num_tfs, N_p), dtype='float64')

        for i in range(num_tfs): # Test each tf individually
            mask = np.zeros((num_replicates, num_tfs, 1), dtype='float64')
            hyp_mask = np.zeros((num_tfs,), dtype='float64')
            hyp_mask[i] = 1
            mask[:, i] = 1
            test_state = (1-mask) * current_state[0] + mask * new_state

            new_prob = self.calc_prob_fn(test_state, new_hyp, old_hyp)
            old_prob = self.calc_prob_fn(current_state[0], old_hyp, new_hyp) #previous_kernel_results.target_log_prob
            is_accepted = self.metropolis_is_accepted(new_prob, old_prob)
            # tf.print(new_prob, old_prob, is_accepted)
            if not is_accepted[0]:
                new_state = (1-mask) * new_state + mask * current_state[0]
                new_hyp[0] = (1-hyp_mask) * new_hyp[0] + hyp_mask * current_state[1]
                new_hyp[1] = (1-hyp_mask) * new_hyp[1] + hyp_mask * current_state[2]
            # prob = tf.cond(tf.equal(is_accepted, tf.constant(True)), lambda:new_prob, lambda:old_prob)

            # new_state = tf.cond(tf.equal(is_accepted, tf.constant(False)),
            #                     lambda:current_state[0], lambda:new_state)
        # new_params = tf.cond(tf.equal(is_accepted, tf.constant(False)),
        #                         lambda:[current_state[1], current_state[2]], lambda:[v, l2])
        return [[new_state, *new_hyp]], f64(0), is_accepted[0]

    def _joint_one_step(self, current_state, previous_kernel_results):
        # Untransformed tf mRNA vectors F (Step 1)
        current_state = current_state[0]

        new_state = tf.identity(current_state[0])
        new_params = []
        S = tf.linalg.diag(self.step_size)
        # MH
        m, K = self.fbar_prior_params(current_state[1], current_state[2])
        # Propose new params
        v = self.kernel_selector.proposal(0, current_state[1]).sample()
        l2 = self.kernel_selector.proposal(1, current_state[2]).sample()
        m_, K_ = self.fbar_prior_params(v, l2)

        # Gibbs step
        fbar = new_state
        z_i = tfd.MultivariateNormalDiag(fbar, self.step_size).sample()

        # Compute K_i(K_i + S)^-1 
        Ksuminv = tf.matmul(K, tf.linalg.inv(K+S))
        # Compute chol(K-K(K+S)^-1 K)
        L = jitter_cholesky(K-tf.matmul(Ksuminv, K))
        c_mu = tf.linalg.matvec(Ksuminv, z_i)
        # Compute nu = L^-1 (f-mu)
        invL = tf.linalg.inv(L)
        nu = tf.linalg.matvec(invL, fbar-c_mu)

        Ksuminv = tf.matmul(K_, tf.linalg.inv(K_+S)) 
        L = jitter_cholesky(K_-tf.matmul(K_, Ksuminv))
        c_mu = tf.linalg.matvec(Ksuminv, z_i)
        fstar = tf.linalg.matvec(L, nu) + c_mu

        new_hyp = [v, l2]
        old_hyp = [current_state[1], current_state[2]]
        new_prob = self.calc_prob_fn(fstar, new_hyp, old_hyp)
        old_prob = self.calc_prob_fn(new_state, old_hyp, new_hyp) #previous_kernel_results.target_log_prob

        is_accepted = self.metropolis_is_accepted(new_prob, old_prob)

        prob = tf.cond(tf.equal(is_accepted, tf.constant(True)), lambda:new_prob, lambda:old_prob)


        new_state = tf.cond(tf.equal(is_accepted, tf.constant(False)),
                            lambda:new_state, lambda:fstar)
        new_params = tf.cond(tf.equal(is_accepted, tf.constant(False)),
                                lambda:[current_state[1], current_state[2]], lambda:[v, l2])

        return [[new_state, *new_params]], prob, is_accepted[0]

    def joint_calc_prob(self, fstar, new_hyp, old_hyp):
        new_m_likelihood = self.likelihood_fn(
            latent=[fstar],
        )
        # σ2_f = 1e-6 * tf.ones(fstar.shape[1], dtype='float64')
        # if 'σ2_f' in self.state_indices:
        #     σ2_f = all_states[self.state_indices['σ2_f']]
        # tf.print(self.data_present, tf.equal(self.data_present, True))
        # print(self.data_present, tf.equal(self.data_present, True))
        if self.data_present:
            new_f_likelihood = tf.reduce_sum(self.latent_likelihood_fn(
                                       latent=[fstar]
                                   ))
        else:
            new_f_likelihood = f64(0)

        new_prob = tf.reduce_sum(new_m_likelihood) + new_f_likelihood

        new_prob += tf.reduce_sum(
            self.kernel_selector.proposal(0, new_hyp[0]).log_prob(old_hyp[0]) +
            self.kernel_selector.proposal(1, new_hyp[1]).log_prob(old_hyp[1])
        )
        if self.kernel_exponential:
            new_hyp = [tf.exp(new_hyp[0]), tf.exp(new_hyp[1])]

        new_prob += tf.reduce_sum(
            self.kernel_priors[0].log_prob(new_hyp[0]) +
            self.kernel_priors[1].log_prob(new_hyp[1])
        )
        return new_prob

    def bootstrap_results(self, init_state):
        init_state = init_state[0]
        prob = self.calc_prob_fn(
            init_state[0], [init_state[1], init_state[2]],
            [init_state[1], init_state[2]])

        return GenericResults(prob, True)
    
    def is_calibrated(self):
        return True



import collections
from numpy import float64 as f64

GenericResults = collections.namedtuple('GenericResults', [
    'target_log_prob',
    'is_accepted',
    'acc_iter',
], defaults=[(f64(0), f64(0))])

MixedKernelResults = collections.namedtuple('MixedKernelResults', [
    'inner_results',
#     'grads_target_log_prob',
#     'step_size',
#     'log_accept_ratio',
    'is_accepted',
    'iteration',
])



import tensorflow as tf
from tensorflow import math as tfm
import tensorflow_probability as tfp

from alfi.mcmc.results import GenericResults
from .mixins import ParamGroupMixin
import numpy as np


class DelaySampler(tfp.mcmc.TransitionKernel, ParamGroupMixin):
    """
    Parameters:
        param: the associated Parameter (only one, not a list)
    """
    def __init__(self, likelihood, param, lower, upper, start_iteration=1):
        super().__init__()
        self.param_group = [param]
        self.likelihood = likelihood
        self.lower = lower
        self.upper = upper
        self.prior = param.prior
        self.start_iteration = start_iteration
        
    def one_step(self, current_state, previous_kernel_results, all_states):
        iteration_number = previous_kernel_results.target_log_prob[0] #just roll with it

        def proceed():
            num_tfs = current_state.shape[0]
            new_state = current_state
            Δrange = np.arange(self.lower, self.upper+1, dtype='float64')
            Δrange_tf = tf.range(self.lower, self.upper+1, dtype='float64')
            for i in range(num_tfs):
                # Generate normalised cumulative distribution
                probs = list()
                mask = np.zeros((num_tfs, ), dtype='float64')
                mask[i] = 1
                
                for Δ in Δrange:
                    test_state = (1-mask) * new_state + mask * Δ

                    probs.append(tf.reduce_sum(self.likelihood(
                        Δ=test_state,
                    )) + tf.reduce_sum(self.prior.log_prob(Δ)))
                # curri = tf.cast(current_state[i], 'int64')
                # start_index = tf.reduce_max([self.lower, curri-2])
                # probs = tf.gather(probs, tf.range(start_index, 
                #                                   tf.reduce_min([self.upper+1, curri+3])))

                probs =  tf.stack(probs) - tfm.reduce_max(probs)
                probs = tfm.exp(probs)
                probs = probs / tfm.reduce_sum(probs)
                cumsum = tfm.cumsum(probs)
                u = tf.random.uniform([], dtype='float64')
                index = tf.where(cumsum == tf.reduce_min(cumsum[(cumsum - u) > 0]))
                chosen = Δrange_tf[index[0][0]]
                new_state = (1-mask) * new_state + mask * chosen
            return new_state
#         tf.print('final chosen state', new_state)
        new_state = tf.cond(iteration_number < self.start_iteration, lambda: current_state, lambda: proceed())
        return new_state, GenericResults([iteration_number+1], True)

    def bootstrap_results(self, init_state, all_states):

        return GenericResults([0], True)
    
    def is_calibrated(self):
        return True




import torch
from torch.nn import Parameter
from torch.distributions.utils import _standard_normal
from gpytorch.constraints import Positive, Interval
from gpytorch.distributions import MultivariateNormal, MultitaskMultivariateNormal
from gpytorch.lazy import DiagLazyTensor

from alfi.configuration import VariationalConfiguration
from alfi.models import OrdinaryLFM, OrdinaryLFMNoPrecompute
from alfi.utilities.torch import softplus, inv_softplus


class LotkaVolterra(OrdinaryLFM):
    """Outputs are predator. Latents are prey"""

    def __init__(self, num_outputs, gp_model, config: VariationalConfiguration, **kwargs):
        super().__init__(num_outputs, gp_model, config, **kwargs)
        self.positivity = Positive()
        self.decay_constraint = Interval(0., 1.5)
        self.raw_decay = Parameter(
            self.positivity.inverse_transform(torch.ones(torch.Size([self.num_outputs, 1]), dtype=torch.float64)))
        self.raw_growth = Parameter(self.positivity.inverse_transform(
            0.5 * torch.ones(torch.Size([self.num_outputs, 1]), dtype=torch.float64)))
        self.raw_initial = Parameter(self.decay_constraint.inverse_transform(
            0.3 + torch.zeros(torch.Size([self.num_outputs, 1]), dtype=torch.float64)))
        # self.true_f = dataset.prey[::3].unsqueeze(0).repeat(self.config.num_samples, 1).unsqueeze(1)

    @property
    def decay_rate(self):
        return self.decay_constraint.transform(self.raw_decay)

    @decay_rate.setter
    def decay_rate(self, value):
        self.raw_decay = self.decay_constraint.inverse_transform(value)

    @property
    def growth_rate(self):
        return softplus(self.raw_growth)

    @growth_rate.setter
    def growth_rate(self, value):
        self.raw_growth = inv_softplus(value)

    @property
    def initial_state(self):
        return softplus(self.raw_initial)

    @initial_state.setter
    def initial_state(self, value):
        self.raw_initial = inv_softplus(value)

    def odefunc(self, t, h):
        """h is of shape (num_samples, num_outputs, 1)"""
        self.nfe += 1
        # if (self.nfe % 100) == 0:
        # print(t, self.t_index, self.f.shape)
        # f shape (num_samples, num_outputs, num_times)
        f = self.latent_gp[:, :, self.t_index].unsqueeze(2)
        if t > self.last_t:
            self.t_index += 1
        self.last_t = t

        dh = self.growth_rate * h * f - self.decay_rate * h
        return dh

    def mix(self, f):
        return softplus(f).repeat(1, self.num_outputs, 1)


class LotkaVolterraState(OrdinaryLFMNoPrecompute):

    def __init__(self, *args, nonzero_mask=None, initial_state=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.nonlinearity = softplus
        self.nonzero_mask = nonzero_mask
        # self.num_data = num_data
        if initial_state is None:
            self.initial_state = torch.tensor([0, 0], dtype=torch.float)
        else:
            self.initial_state = initial_state

    def odefunc(self, t, h, return_mean=False, **kwargs):
        """
        Parameters:
            h shape  (S, D)
        """
        self.nfe += 1
        # Sample from the GP
        # dh = self.gp_model(h).rsample(torch.Size([1]))[0]
        dh = self.gp_model(h)
        # reparam trick:
        mean = dh.mean
        if return_mean:
            return mean
        var = dh.variance
        z = _standard_normal(mean.shape, dtype=mean.dtype, device=mean.device)
        dh = mean + z * (var + 1e-7).sqrt()  # shape (S, D)
        return dh

    def build_output_distribution(self, t, h_samples) -> MultitaskMultivariateNormal:
        """
        Parameters:
            h_samples: shape (T, S, D)
        """
        h_mean = h_samples.mean(dim=1).transpose(0, 1)  # shape was (#outputs, #T, 1)
        h_var = h_samples.var(dim=1).transpose(0, 1) + 1e-7
        # h_mean shape: (D, T)
        self.current_trajectory = h_mean

        if self.config.latent_data_present:
            # todo: make this
            f = self.gp_model(self.timepoint_choices).rsample(torch.Size([self.config.num_samples])).permute(0, 2, 1)
            f = self.nonlinearity(f)
            f_mean = f.mean(dim=0)
            f_var = f.var(dim=0) + 1e-7
            filler = torch.zeros(f_mean.shape[0], h_mean.shape[1]-self.timepoint_choices.shape[0])
            f_mean = torch.cat([f_mean, filler], dim=1)
            f_var = torch.cat([f_var, filler], dim=1)
            h_mean = torch.cat([h_mean, f_mean], dim=0)
            h_var = torch.cat([h_var, f_var], dim=0)

        h_covar = DiagLazyTensor(h_var)
        batch_mvn = MultivariateNormal(h_mean, h_covar)
        return MultitaskMultivariateNormal.from_batch_mvn(batch_mvn, task_dim=0)

    def predict_f(self, t_predict, **kwargs):
        raise NotImplementedError('This model has no latent forces')



import operator
import torch

from functools import reduce
from torch.distributions import Normal
from torch.nn.functional import softplus

from alfi.nn import SimpleBlock1d, SimpleBlock2d, NNBlock1d
from alfi.models import LFM


class NeuralOperator(LFM):
    def __init__(self,
                 block_dim,
                 in_channels,
                 out_channels,
                 modes,
                 width,
                 num_layers=4,
                 params=True,
                 type='fourier'):
        """
        Neural network with different transforms.

        :param block_dim: dimension of the convolutional block
        :param in_channels: input channels
        :param out_channels: output channels
        :param modes: number of Fourier modes or dim of transform mid layers
        :param width: middle channels
        :param num_layers:
        :param params:
        :param type: one of (fourier, mlp)
        """
        super().__init__()
        self.num_outputs = 1
        self.params = params
        if block_dim == 1:
            conv_block = SimpleBlock1d if type == 'fourier' else NNBlock1d
            self.conv = conv_block(in_channels, out_channels, modes, width,
                                      num_layers=num_layers, params=params)
        elif block_dim == 2:
            conv_block = SimpleBlock2d if type == 'fourier' else NNBlock1d
            self.conv = conv_block(in_channels, out_channels, modes, modes, width, num_layers=num_layers,
                                      params=params)
        else:
            pass

    def forward(self, x):
        x = self.conv(x)
        return x#.squeeze()

    def count_params(self):
        c = 0
        for p in self.parameters():
            c += reduce(operator.mul, list(p.size()))
        return c

    def predict_f(self, tx_predict):
        with torch.no_grad():
            y_pred, params_out = self(tx_predict)
            p_y_mean = y_pred[..., 0]
            p_y_var = softplus(y_pred[..., 1])
            p_y_pred = Normal(p_y_mean, p_y_var)
            return p_y_pred, params_out

    def save(self, filepath):
        torch.save(self.state_dict(), str(filepath) + 'lfo.pt')



import torch
import numpy as np

from alfi.utilities.torch import discretisation_length
from . import LFMDataset
from tqdm import tqdm
from torchdiffeq import odeint


class DeterministicLotkaVolterra(LFMDataset):
    """
    Dataset of Lotka-Voltera time series.
      Populations (u,v) evolve according to
        u' = \alpha u - \beta u v
        v' = \delta uv - \gamma v
      with the dataset sampled either with (u_0, v_0) fixed and (\alpha, \beta,
      \gamma, \delta) varied, or varying the initial populations for a fixed set
      of greeks.
    If initial values for (u,v) are provided then the greeks are sampled from
        (0.9,0.05,1.25,0.5) to (1.1,0.15,1.75,1.0)
    if values are provided for the greeks then (u_0 = v_0) is sampled from
        (0.5) to (2.0)
    if both are provided, defaults to initial population mode (greeks vary)
    ----------
    initial_u   : int
        fixed initial value for u
    initial_v   : int
        fixed initial value for v
    fixed_alpha : int
        fixed initial value for \alpha
    fixed_beta  : int
        fixed initial value for \beta
    fixed_gamma : int
        fixed initial value for \gamme
    fixed_delta : int
        fixed initial value for \delta
    end_time : float
        the final time (simulation runs from 0 to end_time)
    steps : int
        how many time steps to take from 0 to end_time
    """
    def __init__(self, initial_u=None, initial_v=None,
                 alpha=None, beta=None, gamma=None, delta=None,
                 steps=13, end_time=12, num_disc=7, fixed_initial=None, silent=False):

        if initial_u is None:
            self.mode = 'greek'
            self.alpha = alpha
            self.beta = beta
            self.gamma = gamma
            self.delta = delta
        else:
            self.mode = 'population'
            self.initial_u = initial_u
            self.initial_v = initial_v

        if not silent:
            print(f'Lotka-Voltera is in {self.mode} mode.')
        self.fixed_initial = fixed_initial
        self.steps = steps
        self.end_time = end_time
        self.num_disc = num_disc
        # Generate data
        self.data = []
        if not silent:
            print("Creating dataset...", flush=True)

        removed = 0
        times, states = self.generate_ts()
        #normalise times
        times = times
        self.times = times

        if self.mode == 'population':
            states = states / 100
        #states = torch.cat((states, times), dim=-1)
        self.prey = states[:, 0]
        self.predator = states[:, 1]
        self.data.append((times[::self.num_disc+1], states[::self.num_disc+1, 1]))

    def generate_ts(self):
        if self.mode == 'population':
            X_0 = np.array([self.initial_u, self.initial_v])
            a = np.random.uniform(0.9, 1.1)
            b = np.random.uniform(0.05, 0.15)
            c = np.random.uniform(1.25, 1.75)
            d = np.random.uniform(0.5, 1.0)
        else:
            equal_pop = self.fixed_initial
            if equal_pop is None:
                while True:
                    equal_pop = np.random.uniform(0.15, 1.)
                    if equal_pop < 0.35 or equal_pop > 0.6:
                        break

            X_0 = np.array([2*equal_pop,equal_pop])
            a, b, c, d = self.alpha, self.beta, self.gamma, self.delta

        def dX_dt(t, X):
            """ Return the growth rate of fox and rabbit populations. """
            return torch.stack([a * X[0] - b * X[0] * X[1],
                               -c * X[1] + d * X[0] * X[1]])

        t = torch.linspace(0, self.end_time, discretisation_length(self.steps, self.num_disc))

        X = odeint(dX_dt, torch.tensor(X_0), t, method='rk4', options=dict(step_size=5e-2))
        return t, X

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return 1



import tensorflow as tf
import tensorflow_probability as tfp

from alfi.mcmc.results import MixedKernelResults
from alfi.utilities.tf import prog

from datetime import datetime


class MixedSampler(tfp.mcmc.TransitionKernel):
    def __init__(self, subsamplers, iteration_callback=None, skip=None):
        """

        Parameters:
            iteration_callback: Callable triggered at the beginning of each iteration if your model needs
             access to the current_state to prepare the likelihood function, for example.
            skip: a boolean array of size |samplers| indicating whether a kernel should be ignored. For debugging.
                        @param data: a tuple (m, f) of shapes (reps, num, time), where time is a tuple (t, τ, common_indices)
        """
        self.iteration_callback = iteration_callback
        self.subsamplers = subsamplers
        self.num_samplers = len(subsamplers)
        # self.one_step_receives_state = [len(signature(k.one_step).parameters) > 2 for k in samplers]
        self.skip = skip
        self.samples = None
        self.initial_state = list()
        self.ordered_param_names = list()
        self.param_transforms = dict()

        super().__init__()

    def before_iteration(self, current_state):
        """This function takes the state list and maps the components to a dictionary"""
        if self.iteration_callback is not None:
            # print()
            # print(current_state)
            parameter_state = dict()
            for i, param_keyset in enumerate(self.ordered_param_names):
                parameter_state.update({
                    param_key: self.param_transforms[param_key](current_state[i][j])
                    for j, param_key in enumerate(param_keyset)
                })

            self.iteration_callback(parameter_state)

    def initialise_state(self):
        if self.samples is None:
            for subsampler in self.subsamplers:
                param_state = [(param.value) for param in subsampler.param_group]
                self.initial_state.append(param_state)
                self.ordered_param_names.append([param.name for param in subsampler.param_group])
                self.param_transforms.update({param.name: param.transform for param in subsampler.param_group})
        else:
            new_state = list()
            for sample_group in self.samples:
                new_group = list()
                for element in sample_group:
                    if type(element) is list:
                        new_group.append([e[-1] for e in element])
                    else:
                        new_group.append(element[-1])
                new_state.append(new_group)
            self.initial_state = new_state

    def sample(self, T=2000, store_every=10, burn_in=1000, report_every=100, skip=None, num_chains=4,
               profile=False):
        print('----- Sampling Begins -----')
        self.T = T
        self.initialise_state()
        current_state = self.initial_state

        def trace_fn(a, previous_kernel_results):
            return previous_kernel_results.is_accepted

        # Run the chain (with burn-in).
        @tf.function
        def run_chain():
            # Run the chain (with burn-in).
            samples, is_accepted = tfp.mcmc.sample_chain(
                num_results=T,
                num_burnin_steps=burn_in,
                current_state=current_state,
                kernel=self,
                trace_fn=trace_fn)

            return samples, is_accepted

        if profile:
            stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
            logdir = '.\\logs\\alfi\\%s' % stamp
            tf.profiler.experimental.start(logdir)
        samples, is_accepted = run_chain()
        if profile:
            tf.profiler.experimental.stop()

        # if self.samples is not None:
        #     for index, sampler in enumerate(self.subsamplers):
        #         param_samples = samples[index]
        #         if type(param_samples[0]) is list:
        #             param_samples = param_samples[0]
        #             for i in range(len(param_samples)):
        #                 self.samples[index][0][i] = tf.concat([self.samples[index][i], param_samples[i]], axis=0)
        #         else:
        #             for i in range(len(param_samples)):
        #                 self.samples[index][i] = tf.concat([self.samples[index][i], param_samples[i]], axis=0)
        #
        #         # param_samples = [[param_samples[i][-1] for i in range(len(param_samples))]]
        #         # param.value = param_samples[-1]
        # else:
        self.samples = samples
        self.is_accepted = is_accepted
        print()
        print('----- Finished -----')

    def one_step(self, current_state, previous_kernel_results):
        # if previous_kernel_results.iteration % 10:
        prog(self.T, previous_kernel_results.iteration)
        # tf.print('running', current_state)
        new_state = list()
        is_accepted = list()
        inner_results = list()
        # Update initial state in case the sampler is run multiple times.
        self.initial_state = current_state
        self.before_iteration(current_state)
        for i in range(self.num_samplers):
            if self.skip is not None and self.skip[i]:
                is_accepted.append(previous_kernel_results.inner_results[i].is_accepted)
                new_state.append(current_state[i])
                inner_results.append(previous_kernel_results.inner_results[i])
                continue

            if hasattr(self.subsamplers[i], 'target_log_prob_fn'):
                tgt_prob = self.subsamplers[i].target_log_prob_fn(*current_state[i])
                if hasattr(previous_kernel_results.inner_results[i], 'accepted_results'):
                    propres = previous_kernel_results.inner_results[i].accepted_results._replace(
                        target_log_prob=tgt_prob)
                    previous_kernel_results.inner_results[i] = previous_kernel_results.inner_results[i]._replace(
                        proposed_results=propres)
                else:

                    previous_kernel_results.inner_results[i] = previous_kernel_results.inner_results[i]._replace(
                        target_log_prob=tgt_prob)

            try:
                # state_chained = tf.expand_dims(current_state[i], 0)
                # print(state_chained)
                result_state, kernel_results = self.subsamplers[i].one_step(
                    current_state[i], previous_kernel_results.inner_results[i])
            except Exception as e:
                tf.print('Failed at ', i, self.subsamplers[i], current_state)
                raise e

            if type(result_state[0]) is list:
                new_state.append([[tf.identity(res) for res in result_state[0]]])
            else:
                new_state.append([tf.identity(res) for res in result_state])

            is_accepted.append(kernel_results.is_accepted)
            inner_results.append(kernel_results)

        return new_state, MixedKernelResults(inner_results, is_accepted, previous_kernel_results.iteration + 1)

    def bootstrap_results(self, init_state):
        """
        """
        self.before_iteration(init_state)

        inner_kernels_bootstraps = list()
        is_accepted = list()
        for i in range(self.num_samplers):
            # self.subsamplers[i].all_states_hack = init_state

            # if self.one_step_receives_state[i]:
            #     results = self.kernels[i].bootstrap_results(init_state[i], init_state)
            #     inner_kernels_bootstraps.append(results)
            #
            # else:
            print('Preparing', self.subsamplers[i])
            results = self.subsamplers[i].bootstrap_results(init_state[i])
            inner_kernels_bootstraps.append(results)

            is_accepted.append(True)

        return MixedKernelResults(inner_kernels_bootstraps, is_accepted, 0)

    def is_calibrated(self):
        return True



from .plotter1d import Plotter1d
from .plotter2d import Plotter2d
from .misc import plot_spatiotemporal_data, plot_phase
from .colours import Colours
tight_kwargs = dict(bbox_inches='tight', pad_inches=0)


__all__ = [
    'Plotter1d',
    'Plotter2d',
    'plot_spatiotemporal_data',
    'plot_phase',
    'Colours',
    'tight_kwargs',
]



import numpy as np

from tensorflow import math as tfm
import tensorflow as tf
from tensorflow import linalg
from tensorflow_probability import distributions as tfd

from datetime import datetime
import pickle
import math

f64 = np.float64
PI = tf.constant(math.pi, dtype='float64')

def discretise(t, num_disc = 10):
    gcd = np.gcd.reduce(t)
    t_augmented = np.arange(0, t[-1]+gcd, gcd)
    N = t_augmented.shape[0]
    
    def calc_N_p(N_p, num_disc):
        '''A helper function to ensure t is a subset of τ'''
        return (N_p-1)*(num_disc+1)+1
    N_p = calc_N_p(N, num_disc)  # Number of time discretisations
    τ = np.linspace(0, t_augmented[-1], N_p, dtype='float64')    # Discretised observation times
    i = int(t[0]/gcd)
    τ = τ[i*num_disc+i:]
    common_indices = np.searchsorted(τ, t)
    return tf.constant(τ), common_indices

def get_time_square(times, N):
    t_1 = tf.transpose(tf.reshape(tf.tile(times, [N]), [N, N]))
    t_2 = tf.reshape(tf.tile(times, [N]), [N, N])
    return t_1, t_2

def add_diag(A, B):
    C= A + tf.linalg.diag(tf.linalg.diag_part(B))
    return C

def logistic(x): # (inverse logit)
    return tfm.exp(x)/(1+tfm.exp(x))

def logit(x, nan_replace=0):
    # print(x>1)
    # if reduce_any(x>1):
    #     return np.inf * ones(x.shape, dtype='float64')
    x = tfm.log(x/(1-x))
    
    x = tf.where(
        tf.math.is_nan(x),
        nan_replace*tf.ones([], x.dtype),
        x)

    return x

def exp(x):
    '''Safe exp'''
    with np.errstate(under='ignore', over='ignore'):
        return np.exp(x)
    
def mult(a, b):
    '''Safe multiplication'''
    with np.errstate(under='ignore', over='ignore', invalid='ignore'):
        c = a*b
        return np.where(np.isnan(c), 0, c)

def jitter_cholesky(A):
    try:
        jitter1 = linalg.diag(1e-7 * tf.ones(A.shape[-1], dtype='float64'))
        return linalg.cholesky(A + jitter1)
    except:
        jitter2 = linalg.diag(1e-5 * tf.ones(A.shape[-1], dtype='float64'))
        return linalg.cholesky(A + jitter2)

def inverse_positivity(fbar):
    return tfm.log(1+tfm.exp(fbar))

def positivity(f_i):
    return tfm.log(tfm.exp(f_i) - 1)

def broadcast_tile(a, h, w):
    x, y = a.shape
    m, n = x * h, y * w
    return tf.reshape(tf.broadcast_to(
        tf.reshape(a, (x, m//(h*x), y, n//(w*y))), (m//h, h, n//w, w)
    ), (m, n))

class ArrayList:
    def __init__(self, shape):
        self.capacity = 100
        self.shape = shape
        self.data = np.zeros((self.capacity, *shape))
        self.size = 0

    def add(self, x):
        if self.size == self.capacity:
            self.capacity *= 4
            newdata = np.zeros((self.capacity, *self.shape))
            newdata[:self.size] = self.data
            self.data = newdata

        self.data[self.size] = x
        self.size += 1

    def get(self):
        data = self.data[:self.size]
        return data

class LogisticNormal():
    def __init__(self, a, b, loc=f64(0), scale=f64(1.5), allow_nan_stats=True, dtype='float64'):
        self.a = f64(a)
        self.b = f64(b)
        self.dist = tfd.LogitNormal(loc, scale, allow_nan_stats=allow_nan_stats)
#         super().__init__(loc, scale, allow_nan_stats=allow_nan_stats)
    def log_prob(self, x):
        x = (x-self.a)/(self.b-self.a)
        log_prob = self.dist.log_prob(x)
        log_prob = tf.where(
            tf.math.is_nan(log_prob),
            -1e2*tf.ones([], log_prob.dtype),
            log_prob)

        return log_prob

def rotate(matrix, shifts):
    """"requested rotate function - assumes matrix shape is mxn and shifts shape is m"""

    # get shape of the input matrix
    shape = tf.shape(matrix)

    # compute and stack the meshgrid to get the index matrix of shape (2,m,n)
    ind = tf.stack(tf.meshgrid(tf.range(shape[0]), tf.range(shape[1]), indexing='ij'))
    # reshape it to (m,n,2)
    ind = tf.transpose(ind, [1,2,0])

    # add the value from shifts to the corresponding row and devide modulo shape[1]
    # this will effectively introduce the desired shift, but at the level of indices
    shifted_ind = tfm.mod(tf.transpose(tf.transpose(ind[:,:,1]) + shifts), shape[1])

    # convert the shifted indices to the right shape
    new_ind = tf.transpose(tf.stack([ind[:,:,0], shifted_ind]) , [1,2,0]) 

    # return the resliced tensor
    return tf.gather_nd(matrix, new_ind)

@tf.function
def prog(T, current):
    mult = 20
    pc = current/T
    x = tf.strings.reduce_join(tf.repeat('=', tf.cast(mult*pc, 'int32')))
    y = tf.strings.reduce_join(tf.repeat('-', tf.cast(mult*(1-pc), 'int32')))
    tf.print('Progress: \t ', tf.round(pc *100.), '%\t| ', x, y, '| ', '\r', sep='', end='')

def save_object(obj, filename):
    stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
    with open(f'saved_models/{filename}-{stamp}.pkl', 'wb') as output:
        pickle.dump(obj, output, pickle.HIGHEST_PROTOCOL)

class FixedDistribution():
    def __init__(self, fixed_value):
        self.fixed_value = fixed_value
    
    def sample(self):
        return self.fixed_value
    
    def log_prob(self, x):
        return tf.constant(0, dtype='float64')


import torch
import numpy as np

from collections.abc import Iterable
from abc import abstractmethod
from typing import List
from torch.utils.data.dataloader import DataLoader
from torch.utils.data import Subset
from torch.nn import Module
from time import time
from pathlib import Path


class Trainer:
    """
    An abstract LFM trainer. Subclasses must implement the `single_epoch` function.

    Parameters
    ----------
    model: The Model.
    optimizers: list of `torch.optim.Optimizer`s. For when natural gradients are used for variational models.
    dataset: Dataset where t_observed (D, T), m_observed (J, T).
    batch_size:
    valid_size: int or float for number of training points or proportion of training points respectively
    test_size: int or float for number of or proportion of training points respectively
    track_parameters: the keys into `named_parameters()` of parameters that the trainer should track. The
                      tracked parameters can be accessed from `parameter_trace`
    train_mask: boolean mask
    """
    def __init__(self,
                 model: Module,
                 optimizers: List[torch.optim.Optimizer],
                 dataset,
                 batch_size=1,
                 valid_size=0.,
                 test_size=0.,
                 device_id=None,
                 track_parameters=None,
                 train_mask=None,
                 checkpoint_dir=None):
        self.model = model
        self.num_epochs = 0
        self.optimizers = optimizers
        self.dataset = dataset
        self.use_natural_gradient = len(self.optimizers) > 1
        self.batch_size = batch_size

        # Dataset splits
        dataset_size = len(dataset)
        if isinstance(valid_size, int):
            valid_size /= dataset_size
        if isinstance(test_size, int):
            test_size /= test_size
        indices = np.arange(dataset_size)
        np.random.shuffle(indices)
        valid_split = int(np.floor(valid_size * dataset_size))
        test_split = int(np.floor(test_size * dataset_size))
        # TODO this was added as a hack, for unknown reasons
        duplicate = lambda ind: np.concatenate([ind, ind + dataset_size])
        # self.valid_indices = duplicate(indices[:valid_split])
        # self.test_indices = duplicate(indices[valid_split:test_split + valid_split])
        # self.train_indices = duplicate(indices[test_split + valid_split:])
        self.valid_indices = indices[:valid_split]
        self.test_indices = indices[valid_split:test_split + valid_split]
        self.train_indices = indices[test_split + valid_split:]
        self.data_loader = self.test_loader = self.valid_loader = None
        self.set_loaders()

        self.losses = None
        self.train_mask = train_mask
        self.checkpoint_dir = checkpoint_dir
        self.parameter_trace = None
        self.device_id = device_id

        if track_parameters is not None:
            named_params = dict(model.named_parameters())
            self.parameter_trace = {key: [named_params[key].detach()] for key in track_parameters}

    def set_loaders(self):
        valid_data = Subset(self.dataset, self.valid_indices)
        test_data = Subset(self.dataset, self.test_indices)
        train_data = Subset(self.dataset, self.train_indices)
        self.test_loader = DataLoader(test_data, batch_size=self.batch_size, shuffle=False)
        self.valid_loader = DataLoader(valid_data, batch_size=self.batch_size, shuffle=False)
        self.data_loader = DataLoader(train_data, batch_size=self.batch_size, shuffle=False)

    def train(self, epochs=20, report_interval=1, reporter_callback=None, **kwargs):
        """
        Parameters:
            epochs: the number of training epochs to run
            report_interval: every report_interval epochs, a progress string will be printed
            reporter_callback: function called every report_interval
        """
        self.model.train()

        losses = list()
        times = list()
        end_epoch = self.num_epochs+epochs

        for epoch in range(epochs):
            epoch_loss = self.single_epoch(epoch=self.num_epochs, **kwargs)
            t = time()
            times.append((t, epoch_loss))

            if isinstance(epoch_loss, Iterable):
                epoch_loss, loss_breakdown = epoch_loss
            else:
                loss_breakdown = [epoch_loss]

            if (epoch % report_interval) == 0:
                if reporter_callback is not None:
                    reporter_callback(self.num_epochs)
                print('Epoch %03d/%03d - Loss: %.2f ' % (
                    self.num_epochs + 1, end_epoch, epoch_loss), end='')
                if len(loss_breakdown) > 1:
                    print(' '.join(map(lambda l: '%.2f' % l, loss_breakdown[1:])), end='')

                self.print_extra()
                if self.checkpoint_dir is not None:
                    self.model.save(self.checkpoint_dir / f'epoch{epoch}')
            losses.append(loss_breakdown)

            self.after_epoch()
            self.num_epochs += 1

        losses = torch.tensor(losses).cpu().numpy()
        if self.losses is None:
            self.losses = np.empty((0, losses.shape[1]))
        self.losses = np.concatenate([self.losses, losses], axis=0)
        return times

    @abstractmethod
    def single_epoch(self, epoch=0, **kwargs):
        raise NotImplementedError

    def set_optimizers(self, optimizers):
        self.optimizers = optimizers

    def print_extra(self):
        print('')

    def after_epoch(self):
        if self.parameter_trace is not None:
            params = dict(self.model.named_parameters())
            for key in params:
                if key in self.parameter_trace:
                    self.parameter_trace[key].append(params[key].detach().clone())

    def save(self, folder_prefix='model', save_model=True, save_losses=True, save_optimizer=False, additional=None):
        """
        Save the model and trainer information.

        :param folder_prefix: the folder name containing the saved files will be folder_prefix-timestamp
        :param save_model:
        :param save_losses:
        :param save_optimizer:
        :param additional: object or list of objects to save alongside the default items.
        :return:
        """
        timestamp = int(time())
        path = Path(f'./{folder_prefix}-{timestamp}/')
        if path.exists():
            raise IOError('Path seems to already exist.')
        path.mkdir()

        torch.save((self.train_indices, self.test_indices, self.valid_indices), path / 'indices.pt')
        if save_model:
            torch.save(self.model.state_dict(), path / 'model.pt')

        if save_losses:
            torch.save(self.losses, path / 'losses.pt')

        if save_optimizer:
            for i, optimizer in enumerate(self.optimizers):
                torch.save(optimizer.state_dict(), path / 'optim[{i}].pt')

        if additional is not None:
            torch.save(additional, path / 'additional.pt')
        return timestamp

    def load(self, file_prefix='model', timestamp=''):
        """
        Loads optionally the model state dict, losses array, optimizer state dict, in that order.

        :param file_prefix:
        :param timestamp:
        """
        path = Path(f'./{file_prefix}-{timestamp}/')
        print('Loading from', path)
        if (path / 'indices.pt').exists():
            indices = torch.load(path / 'indices.pt')
            self.train_indices, self.test_indices, self.valid_indices = indices
            self.set_loaders()
        if (path / 'model.pt').exists():
            state_dict = torch.load(path / 'model.pt', map_location=torch.device('cpu'))
            self.model.load_state_dict(state_dict)
        if (path / 'losses.pt').exists():
            self.losses = torch.load(path / 'losses.pt')
        for i, optim in enumerate([f'optim[{i}].pt' for i in range(len(self.optimizers))]):
            if (path / optim).exists():
                self.optimizers[i].load_state_dict(torch.load(path / optim))



from alfi.mcmc.samplers.mixed import MixedSampler
from alfi.mcmc.samplers.mh import MetropolisKernel
from alfi.mcmc.samplers.delay import DelaySampler
from alfi.mcmc.samplers.latent import LatentGPSampler
from alfi.mcmc.samplers.gibbs import GibbsSampler
from alfi.mcmc.samplers.hmc import HMCSampler


__all__ = [
    'MixedSampler',
    'LatentGPSampler',
    'HMCSampler',
    'DelaySampler',
    'GibbsSampler',
    'MetropolisKernel',
]


import torch
from torch.nn import Parameter, Module
from alfi.utilities.torch import compl_mul1d, compl_mul2d


"""
Original Authors: Li, Z., Kovachki, N., Azizzadenesheli, K., Liu, B., Bhattacharya, K., Stuart, A. and Anandkumar, A., 2020. 
Fourier neural operator for parametric partial differential equations. 
arXiv preprint arXiv:2010.08895.

Edited: Moss, Jacob. 2021
- fixes to work with Torch 1.9.0 (complex dtypes)
"""
class SpectralConv1d(Module):
    def __init__(self, in_channels, out_channels, modes1):
        super(SpectralConv1d, self).__init__()
        """
        1D Fourier layer. It does FFT, linear transform, and Inverse FFT.
        """
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.modes1 = modes1  #Number of Fourier modes to multiply, at most floor(N/2) + 1

        self.scale = (1 / (in_channels*out_channels))
        self.weights1 = Parameter(
            self.scale * torch.rand(in_channels, out_channels, self.modes1, dtype=torch.complex64))

    def forward(self, x):
        batchsize = x.shape[0]
        #Compute Fourier coeffcients up to factor of e^(- something constant)
        x_ft = torch.fft.rfft(x)

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(batchsize, self.in_channels, x.size(-1)//2 + 1,
                             device=x.device, dtype=torch.complex64)
        out_ft[:, :, :self.modes1] = compl_mul1d(x_ft[:, :, :self.modes1], self.weights1)

        #Return to physical space
        x = torch.fft.irfft(out_ft, n=x.size(-1))
        return x, out_ft


class SpectralConv2d(Module):
    def __init__(self, in_channels, out_channels, modes1, modes2):
        super(SpectralConv2d, self).__init__()

        """
        2D Fourier layer. It does FFT, linear transform, and Inverse FFT.
        """

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.modes1 = modes1 #Number of Fourier modes to multiply, at most floor(N/2) + 1
        self.modes2 = modes2

        self.scale = (1 / (in_channels * out_channels))
        self.weights1 = Parameter(
            self.scale * torch.rand(in_channels, out_channels, self.modes1, self.modes2, dtype=torch.complex64))
        self.weights2 = Parameter(
            self.scale * torch.rand(in_channels, out_channels, self.modes1, self.modes2, dtype=torch.complex64))

    def forward(self, x):
        batchsize = x.shape[0]
        #Compute Fourier coeffcients up to factor of e^(- something constant)
        x_ft = torch.fft.rfft2(x)

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(batchsize, self.in_channels,  x.size(-2), x.size(-1)//2 + 1,
                             device=x.device, dtype=torch.complex64)
        out_ft[:, :, :self.modes1, :self.modes2] = compl_mul2d(x_ft[:, :, :self.modes1, :self.modes2], self.weights1)
        out_ft[:, :, -self.modes1:, :self.modes2] = compl_mul2d(x_ft[:, :, -self.modes1:, :self.modes2], self.weights2)
        out_ft[:, :, :self.modes1, -self.modes2:] = compl_mul2d(x_ft[:, :, :self.modes1, -self.modes2:], self.weights2)

        #Return to physical space
        x = torch.fft.irfft2(out_ft, s=(x.size(-2), x.size(-1)))
        return x, out_ft



from .spectral_conv import SpectralConv1d
from .conv_block import SimpleBlock1d, SimpleBlock2d
from .lp_loss import LpLoss
from .encoders import MuSigmaEncoder
from .conv_block_nn import NNBlock1d


__all__ = [
    'SimpleBlock1d',
    'SimpleBlock2d',
    'SpectralConv1d',
    'LpLoss',
    'MuSigmaEncoder',
    'NNBlock1d',
]



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch

from .base_plotter import Plotter
from alfi.datasets import scaled_barenco_data
from alfi.models import VariationalLFM, TrainMode, OrdinaryLFMNoPrecompute
from .colours import Colours
from alfi.utilities import is_cuda


plt.style.use('seaborn')
plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.serif'] = 'CMU Serif'
sns.set(style='white', font="CMU Serif")


class Plotter2d(Plotter):
    """
    This Plotter is designed for 2D LFMs or LFOs.
    """

    def __init__(self, model, output_names, style='seaborn'):
        super().__init__(model, output_names, style=style)
        self.num_outputs = len(self.output_names)

    def plot_vector_gp(self, h, true_h,
                       ax=None,
                       figsize=(3, 3),
                       show=True,
                       save_name=None,
                       independent=True,
                       batch_index=None,
                       title='',
                       labels=None,
                       scat_kwargs=None,
                       cell_colors=None,
                       plot_inducing=True):
        with torch.no_grad():
            x1, x2 = h.cpu()
            true_x1, true_x2 = true_h.cpu()
            if not show:
                plt.ioff()

            if ax is None:
                plt.figure(figsize=figsize)
                ax = plt.subplot()
            plt.title(title)
            if scat_kwargs is None:
                scat_kwargs = dict(
                    alpha=0.8, s=5)

            # Plot trajectory
            ax.plot(x1, x2, color='red', alpha=0.5, linewidth=1)

            # Plot cell points
            indices = np.intersect1d(true_x1.nonzero(), true_x2.nonzero())

            if cell_colors is None:
                cell_colors = 'black'
            else:
                cell_colors = cell_colors[indices]
                
            main_scatter = ax.scatter(
                true_x1[indices], true_x2[indices],
                cmap='viridis', c=cell_colors, **scat_kwargs)

            if plot_inducing:
                # Plot inducing vectors
                inducing_points = self.model.inducing_points.cuda() if is_cuda() else self.model.inducing_points
                num_inducing = inducing_points.shape[1]
                if isinstance(self.model, OrdinaryLFMNoPrecompute):
                    ax.scatter(self.model.initial_state[:, 0].cpu(), self.model.initial_state[:, 1].cpu())

                    self.model.set_mode(TrainMode.GRADIENT_MATCH)
                    h_grad = self.model((inducing_points, None)).mean
                    if independent:
                        ind_x1 = inducing_points[0, :, 0]
                        ind_x2 = inducing_points[1, :, 0]
                    else:
                        ind_x1 = inducing_points[0, :, 0]
                        ind_x2 = inducing_points[0, :, 1]

                elif batch_index is not None:
                    initial_state = self.model.initial_state.view(2, -1)
                    ax.scatter(initial_state[0], initial_state[1])

                    self.model.set_mode(TrainMode.NORMAL)
                    self.model(inducing_points[0, :, 0])
                    traj = self.model.current_trajectory
                    ind = traj.view(2, -1, num_inducing)[:, batch_index]
                    ind_x1 = ind[0]
                    ind_x2 = ind[1]

                    self.model.set_mode(TrainMode.GRADIENT_MATCH)
                    h_grad = self.model((inducing_points, traj)).mean.view(num_inducing, 2, -1)
                    h_grad = h_grad[..., batch_index]

                grad_x1 = h_grad[:, 0].cpu()
                grad_x2 = h_grad[:, 1].cpu()
                ax.quiver(ind_x1.cpu(), ind_x2.cpu(), grad_x1, grad_x2)

            if labels is not None:
                ax.set_xlabel(labels[0])
                ax.set_ylabel(labels[1])
            if save_name is not None:
                plt.savefig(save_name)
            if not show:
                plt.close(plt.gcf())

        return ax, main_scatter



import torch

from typing import List
from torch.optim.lr_scheduler import StepLR
from torch.optim import Optimizer
from torch.nn.functional import mse_loss
from torch.distributions.kl import kl_divergence

from alfi.nn import LpLoss
from alfi.utilities.torch import is_cuda
from alfi.utilities.data import context_target_split as cts
from alfi.models import NeuralOperator
from alfi.trainers import Trainer


class NeuralOperatorTrainer(Trainer):
    """
    Parameters:
        batch_size: in the case of the transcriptional regulation model, we train the entire gene set as a batch
    """
    def __init__(self,
                 lfm: NeuralOperator,
                 optimizers: List[Optimizer],
                 train_loader,
                 test_loader,
                 warm_variational=-1,
                 **kwargs):
        super().__init__(lfm, optimizers, train_loader.dataset, **kwargs)
        self.warm_variational = warm_variational
        self.optimizer = optimizers[0]
        self.train_loader = train_loader
        self.test_loader = test_loader
        gamma = 0.5
        step_size = 100

        self.scheduler = StepLR(self.optimizer, step_size=step_size, gamma=gamma)

        self.loss_fn = LpLoss(size_average=True)

    def single_epoch(self, step_size=1e-1, epoch=0, num_context=7, num_target=4, **kwargs):
        self.model.train()
        train_mse = 0
        train_l2 = 0
        train_loss = 0
        train_params_mse = 0
        batch_size = self.train_loader.batch_size
        for x, y, params in self.train_loader:
            if is_cuda():
                x, y, params = x.cuda(), y.cuda(), params.cuda()
            x_context, y_context, x_target, y_target = cts(x, y, num_context, num_target)

            x_context = x_context
            y_context = y_context
            x_target = x_target
            y_target = y_target
            self.optimizer.zero_grad()
            # out, params_out = self.model(x_context, y_context, x_target, y_target)
            p_y_pred, y_params, q_target, q_context = self.model(x_context, y_context, x_target, y_target)

            loss = self._loss(p_y_pred, y_target.squeeze(-1), q_target, q_context)
            mse = mse_loss(p_y_pred.mean, y.squeeze(-1), reduction='mean')
            # l2 = self.loss_fn(out, y.view(batch_size, -1))
            params_mse = mse_loss(y_params, params.view(batch_size, -1), reduction='mean')
            # total_loss = l2 + params_mse
            total_loss = loss + params_mse
            total_loss.backward()  # use the l2 relative loss
            self.optimizer.step()

            train_mse += mse.item()
            # train_l2 += l2.item()
            train_loss += total_loss.item()
            train_params_mse += params_mse.item()

        self.scheduler.step()
        self.model.eval()

        test_l2 = 0.0
        test_mse = 0.0
        test_params_mse = 0.0
        test_loss = 0.0
        with torch.no_grad():
            for x, y, params in self.test_loader:
                # x, y = x.cuda(), y.cuda()
                x_context, y_context, x_target, y_target = cts(x, y, num_context, num_target)
                x_context = x_context
                y_context = y_context
                x_target = x_target
                y_target = y_target
                p_y_pred, y_params = self.model(x_context, y_context, x_target, y_target)
                loss = -p_y_pred.log_prob(y_target.squeeze(-1)).mean(dim=0).sum()
                test_mse += mse_loss(p_y_pred.mean, y.squeeze(-1), reduction='mean')
                # test_mse += mse_loss(out[..., 0:1], y, reduction='mean')
                # test_l2 += self.loss_fn(out, y.view(self.test_loader.batch_size, -1)).item()
                params_mse = mse_loss(y_params, params.view(self.test_loader.batch_size, -1), reduction='mean')
                test_params_mse += params_mse
                test_loss += loss + params_mse

        train_mse /= len(self.train_loader)
        train_l2 /= len(self.train_loader)
        train_params_mse /= len(self.train_loader)
        train_loss /= len(self.train_loader)
        test_l2 /= len(self.test_loader)
        test_mse /= len(self.test_loader)
        test_params_mse /= len(self.test_loader)
        test_loss /= len(self.test_loader)

        return train_loss, (test_loss, train_l2, test_l2, train_params_mse, test_params_mse)

    def _loss(self, p_y_pred, y_target, q_target, q_context):
        """
        Computes Neural Process loss.

        Parameters
        ----------
        p_y_pred : one of torch.distributions.Distribution
            Distribution over y output by Neural Process.

        y_target : torch.Tensor
            Shape (batch_size, num_target, y_dim)

        q_target : one of torch.distributions.Distribution
            Latent distribution for target points.

        q_context : one of torch.distributions.Distribution
            Latent distribution for context points.
        """
        # Log likelihood has shape (batch_size, num_target, y_dim). Take mean
        # over batch and sum over number of targets and dimensions of y
        nll = -p_y_pred.log_prob(y_target).mean(dim=0).sum()
        # KL has shape (batch_size, r_dim). Take mean over batch and sum over
        # r_dim (since r_dim is dimension of normal distribution)
        kl = kl_divergence(q_target, q_context).mean(dim=0).sum()
        return nll + kl



from abc import abstractmethod

import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow import math as tfm

from alfi.mcmc.results import GenericResults

import numpy as np
f64 = np.float64


class MetropolisKernel(tfp.mcmc.TransitionKernel):
    def __init__(self, step_size, tune_every=20):
        self.step_size = tf.Variable(step_size)
        self.tune_every = tune_every

    def metropolis_is_accepted(self, new_log_prob, old_log_prob):
        alpha = tf.math.exp(new_log_prob - old_log_prob)
        return tf.random.uniform((1,), dtype='float64') < tf.math.minimum(f64(1), alpha)
    #     if is_tensor(alpha):
    #         alpha = alpha.numpy()
    #     return not np.isnan(alpha) and random.random() < min(1, alpha)

    def one_step(self, current_state, previous_kernel_results):
        new_state, prob, is_accepted = self._one_step(current_state, previous_kernel_results)

        acc_rate, iteration = previous_kernel_results.acc_iter
        acc = acc_rate*iteration
        iteration += f64(1)
        acc_rate = tf.cond(tf.equal(is_accepted, tf.constant(True)), 
                        lambda: (acc+1)/iteration, lambda: acc/iteration)
        # tf.print(acc_rate, iteration)
        tf.cond(tf.equal(tfm.floormod(iteration, self.tune_every), 0), lambda: self.tune(acc_rate), lambda:None)

        return new_state, GenericResults(prob, is_accepted, (acc_rate, iteration)) # TODO for multiple TFs

    @abstractmethod
    def _one_step(self, current_state, previous_kernel_results):
        pass
    
    def tune(self, acc_rate):
        self.step_size.assign(tf.case([
                (acc_rate < 0.01, lambda: self.step_size * 0.1),
                (acc_rate < 0.1, lambda: self.step_size * 0.5),
                (acc_rate < 0.3, lambda: self.step_size * 0.9),
                (acc_rate > 0.95, lambda: self.step_size * 10.0),
                (acc_rate > 0.75, lambda: self.step_size * 2.0),
                (acc_rate > 0.5, lambda: self.step_size * 1.1),
            ],
            default=lambda:self.step_size
        ))
        tf.print('Updating step_size', self.step_size[0], 'due to acc rate', acc_rate)
    
    def is_calibrated(self):
        return True




from .loaders import load_barenco_puma, DataHolder, barenco_params, scaled_barenco_data, load_covid
from .lfm_dataset import LFMDataset
from .toy import ToyTranscriptomicGenerator, ToyTranscriptomics
from .datasets import (
    TranscriptomicTimeSeries,
    P53Data,
    HafnerData,
    DrosophilaSpatialTranscriptomics
)
from .lotkavolterra import DeterministicLotkaVolterra
from .toy_spatial import HomogeneousReactionDiffusion, ReactionDiffusion, ReactionDiffusionGenerator

__all__ = [
    'load_barenco_puma',
    'barenco_params',
    'scaled_barenco_data',
    'load_covid',
    'LFMDataset',
    'P53Data',
    'DataHolder',
    'HafnerData',
    'TranscriptomicTimeSeries',
    'ToyTranscriptomics',
    'HomogeneousReactionDiffusion',
    'ReactionDiffusion',
    'ReactionDiffusionGenerator',
    'DrosophilaSpatialTranscriptomics',
    'DeterministicLotkaVolterra',
]



import torch
import torch.nn.functional as F

from torch.nn import Module, ModuleList, Linear, Conv1d
from .spectral_conv import SpectralConv1d, SpectralConv2d


"""
Authors: Li, Z., Kovachki, N., Azizzadenesheli, K., Liu, B., Bhattacharya, K., Stuart, A. and Anandkumar, A., 2020. 
Fourier neural operator for parametric partial differential equations. 
arXiv preprint arXiv:2010.08895.
"""
class SimpleBlock1d(Module):
    def __init__(self, in_channels, out_channels, modes, width, num_layers=4, params=True):
        super(SimpleBlock1d, self).__init__()

        """
        The overall network. It contains 4 layers of the Fourier layer.
        1. Lift the input to the desire channel dimension by self.fc0 .
        2. 4 layers of the integral operators u' = (W + K)(u).
            W defined by self.w; K defined by self.conv .
        3. Project from the channel space to the output space by self.fc1 and self.fc2 .

        input: the solution of the initial condition and location (a(x), x)
        input shape: (batchsize, x=s, c=2)
        output: the solution of a later timestep
        output shape: (batchsize, x=s, c=1)
        """
        self.params = params
        self.modes1 = modes
        self.width = width
        self.num_layers = num_layers
        self.fc0 = Linear(in_channels, self.width)
        # self.fc0_parameters = Linear(self.width * self.modes1 * num_layers * 2, 100)
        if self.params:
            self.fc0_parameters = Linear((1 + self.modes1 * num_layers*2)* self.width, 50)
            self.fc1_parameters = Linear(50, 5*3)
        self.spectral_layers = ModuleList([
            SpectralConv1d(self.width, self.width, self.modes1)
            for _ in range(num_layers)
        ])
        self.weight_layers = ModuleList([
            Conv1d(self.width, self.width, 1)
            for _ in range(num_layers)
        ])
        self.param_layers = ModuleList([
            Conv1d(self.width, self.width, 1)
            for _ in range(num_layers)
        ])

        self.fc1 = Linear(self.width, 128)
        self.fc2 = Linear(128, out_channels)

    def forward(self, x):
        batchsize = x.shape[0]
        x = self.fc0(x)
        x = x.permute(0, 2, 1)
        out_fts = torch.zeros(batchsize, self.width, self.modes1*self.num_layers,
                              dtype=torch.complex64, device=x.device)
        out_p = torch.zeros(batchsize, self.width, x.shape[-1])
        p = None
        for i in range(self.num_layers):
            conv = self.spectral_layers[i]
            w = self.weight_layers[i]
            w_params = self.param_layers[i]
            x1, out_ft = conv(x)
            out_fts[..., i*self.modes1:(i+1)*self.modes1] = out_ft[..., :self.modes1]
            x2 = w(x)
            p_w = w_params(x)
            if p is None:
                p = p_w
            else:
                p = p_w + p
            x = x1 + x2
            if i < (self.num_layers - 1):
                x = F.relu(x)

        x = x.permute(0, 2, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.fc2(x)

        if self.params:
            out_fts = torch.stack([out_fts.real, out_fts.imag], dim=-1)
            # print(p.mean(-1).unsqueeze(-1).shape, out_fts.view(batchsize, self.width, -1).shape)
            # params = self.fc0_parameters(out_fts.reshape(batchsize, -1))
            # params = F.relu(params)
            # params = self.fc1_parameters(params)
            params = torch.cat([p.mean(-1).unsqueeze(-1), out_fts.view(batchsize, self.width, -1)], dim=-1)
            # print(params.shape)
            # print(self.fc0_parameters)
            # params = p.mean(-1)
            params = self.fc0_parameters(params.view(batchsize, -1))
            params = F.relu(params)
            params = self.fc1_parameters(params)
            return x, params

        return x


class SimpleBlock2d(Module):
    def __init__(self, in_channels, out_channels, modes1, modes2, width, num_layers=4, params=True):
        super(SimpleBlock2d, self).__init__()

        """
        The overall network. It contains 4 layers of the Fourier layer.
        1. Lift the input to the desire channel dimension by self.fc0 .
        2. 4 layers of the integral operators u' = (W + K)(u).
            W defined by self.w; K defined by self.conv .
        3. Project from the channel space to the output space by self.fc1 and self.fc2 .

        input: the solution of the coefficient function and locations (a(x, y), x, y)
        input shape: (batchsize, x=s, y=s, c=in_channels)
        output: the solution
        output shape: (batchsize, x=s, y=s, c=1)
        """
        self.params = params
        self.modes1 = modes1
        self.modes2 = modes2
        self.width = width
        self.num_layers = num_layers
        self.fc0 = Linear(in_channels, self.width)
        if params:
            self.fc0_parameters = Linear(self.width * self.modes1 * self.modes2 * num_layers**2 * 2, 200)
            self.fc1_parameters = Linear(200, 5)
        self.spectral_layers = ModuleList([
            SpectralConv2d(self.width, self.width, self.modes1, self.modes2)
            for _ in range(num_layers)
        ])
        self.weight_layers = ModuleList([
            Conv1d(self.width, self.width, 1)
            for _ in range(num_layers)
        ])

        self.fc1 = Linear(self.width, 128)
        self.fc2 = Linear(128, out_channels)

    def forward(self, x):
        batchsize = x.shape[0]
        size_x, size_y = x.shape[1], x.shape[2]

        x = self.fc0(x)
        x = x.permute(0, 3, 1, 2)
        out_fts = torch.zeros(batchsize, self.width, self.modes1*self.num_layers, self.modes2*self.num_layers,
                              dtype=torch.complex64, device=x.device)

        for i in range(self.num_layers):
            conv = self.spectral_layers[i]
            w = self.weight_layers[i]
            x1, out_ft = conv(x)
            out_fts[..., i*self.modes1:(i+1)*self.modes1, i*self.modes2:(i+1)*self.modes2] = \
                out_ft[..., :self.modes1, :self.modes2]
            x2 = w(x.view(batchsize, self.width, -1)).view(batchsize, self.width, size_x, size_y)
            x = x1 + x2
            if i < (self.num_layers - 1):
                x = F.relu(x)

        x = x.permute(0, 2, 3, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.fc2(x)

        if self.params:
            out_fts = torch.stack([out_fts.real, out_fts.imag], dim=-1)
            params = self.fc0_parameters(out_fts.reshape(batchsize, -1))
            params = F.relu(params)
            params = self.fc1_parameters(params)
            return x, params

        return x



from abc import abstractmethod

import torch
import gpytorch
from torch.distributions import Distribution
from torchdiffeq import odeint
from gpytorch.lazy import DiagLazyTensor

from .variational_lfm import VariationalLFM
from . import TrainMode
from alfi.configuration import VariationalConfiguration
from alfi.utilities.torch import is_cuda


class OrdinaryLFM(VariationalLFM):
    """
    Variational approximation for an LFM based on an ordinary differential equation (ODE).
    Inheriting classes must override the `odefunc` function which encodes the ODE.
    """

    def __init__(self,
                 num_outputs,
                 gp_model,
                 config: VariationalConfiguration,
                 initial_state=None,
                 **kwargs):
        super().__init__(num_outputs, gp_model, config, **kwargs)
        self.nfe = 0
        self.latent_gp = None
        if initial_state is None:
            self.initial_state = torch.zeros(torch.Size([self.num_outputs, 1]), dtype=self.dtype)
        else:
            self.initial_state = initial_state

    @property
    def initial_state(self):
        return self._initial_state

    @initial_state.setter
    def initial_state(self, value):
        value = value.cuda() if is_cuda() else value
        self._initial_state = value

    def forward(self, t, step_size=1e-1, return_samples=False, **kwargs):
        """
        t : torch.Tensor
            Shape (num_times)
        h : torch.Tensor the initial state of the ODE
            Shape (num_genes, 1)
        Returns
        -------
        Returns evolved h across times t.
        Shape (num_genes, num_points).
        """
        self.nfe = 0

        # Get GP outputs
        if self.train_mode == TrainMode.GRADIENT_MATCH:
            t_f = t[0]
            t_output = t_f
            h0 = t[1].unsqueeze(0).repeat(self.config.num_samples, 1, 1)
        else:
            t_f = torch.arange(t.min(), t.max()+step_size/3, step_size/3, device=t.device)
            t_output = t
            h0 = self.initial_state.to(t.device)
            h0 = h0.unsqueeze(0).repeat(self.config.num_samples, 1, 1)

        q_f = self.gp_model(t_f)

        self.latent_gp = q_f.rsample(torch.Size([self.config.num_samples])).permute(0, 2, 1)  # (S, I, T)
        self.latent_gp = self.nonlinearity(self.latent_gp)
        self.latent_gp = self.mix(self.latent_gp)

        if self.train_mode == TrainMode.GRADIENT_MATCH:
            h_samples = self.odefunc(t_f, h0)
            h_samples = h_samples.permute(2, 0, 1)
        else:
            # Integrate forward from the initial positions h0.
            self.t_index = 0
            self.last_t = t_f.min() - 1
            h_samples = odeint(self.odefunc, h0, t, method='rk4', options=dict(step_size=step_size)) # (T, S, num_outputs, 1)

        # self.t_index = None
        # self.last_t = None
        if return_samples:
            return h_samples

        dist = self.build_output_distribution(t_output, h_samples)
        self.latent_gp = None
        return dist

    def build_output_distribution(self, t, h_samples) -> Distribution:
        """
        Parameters:
            h_samples: shape (T, S, D)
        """
        h_mean = h_samples.mean(dim=1).squeeze(-1).transpose(0, 1)  # shape was (#outputs, #T, 1)
        h_var = h_samples.var(dim=1).squeeze(-1).transpose(0, 1) + 1e-7

        # TODO: make distribution something less constraining
        if self.config.latent_data_present:
            # todo: make this
            f = self.gp_model(t).rsample(torch.Size([self.config.num_samples])).permute(0, 2, 1)
            # f = self.nonlinearity(f)
            f_mean = f.mean(dim=0)
            f_var = f.var(dim=0) + 1e-7
            h_mean = torch.cat([h_mean, f_mean], dim=0)
            h_var = torch.cat([h_var, f_var], dim=0)

        h_covar = DiagLazyTensor(h_var)  # (num_tasks, t, t)
        batch_mvn = gpytorch.distributions.MultivariateNormal(h_mean, h_covar)
        return gpytorch.distributions.MultitaskMultivariateNormal.from_batch_mvn(batch_mvn, task_dim=0)

    @abstractmethod
    def odefunc(self, t, h, **kwargs):
        """
        Parameters:
            h: shape (num_samples, num_outputs, 1)
        """
        pass

    def sample_latents(self, t, num_samples=1):
        q_f = self.gp_model(t)
        return self.nonlinearity(q_f.sample(torch.Size([num_samples])))

    def nonlinearity(self, f):
        return f

    def mix(self, f):
        return f#.repeat(1, self.num_outputs, 1)  # (S, I, t)



from abc import ABC

import torch
from torchdiffeq import odeint

from . import OrdinaryLFM, TrainMode
from alfi.configuration import VariationalConfiguration


class OrdinaryLFMNoPrecompute(OrdinaryLFM, ABC):
    """
    Variational approximation for an LFM based on an ordinary differential equation (ODE).
    NoPrecompute means that this class does not precompute the GP at timepoints.
    This enables giving the state as input rather than time.
    Inheriting classes must override the `odefunc` function which encodes the ODE.
    """
    def __init__(self, num_outputs, gp_model, config: VariationalConfiguration, **kwargs):
        super().__init__(num_outputs, gp_model, config, **kwargs)
        self.nfe = 0
        self.f = None

    @OrdinaryLFM.initial_state.getter
    def initial_state(self):
        return self._initial_state.unsqueeze(0)\
                                  .repeat(self.config.num_samples, 1, 1)

    def forward(self, t, step_size=1e-1, return_samples=False, **kwargs):
        """
        t : torch.Tensor
            Shape (num_times)
        h : torch.Tensor the initial state of the ODE
            Shape (num_genes, 1)
        Returns
        -------
        Returns evolved h across times t.
        Shape (num_genes, num_points).
        """
        self.nfe = 0

        # Get GP outputs
        if self.train_mode == TrainMode.GRADIENT_MATCH:
            h0 = t[0]
            t_output = t[1]
            h_samples = self.odefunc(None, h0, return_mean=True).unsqueeze(1)
        else:
            t_output = t
            h0 = self.initial_state

            # Integrate forward from the initial positions h0.
            h_samples = odeint(self.odefunc, h0, t, method='rk4', options=dict(step_size=step_size)) # (T, S, num_outputs, 1)

        if return_samples:
            return h_samples

        dist = self.build_output_distribution(t_output, h_samples)
        self.f = None
        return dist

    def sample_latents(self, t, num_samples=1):
        q_f = self.gp_model(t)
        return self.nonlinearity(q_f.sample(torch.Size([num_samples])))




import torch
import numpy as np
import pandas as pd
from pathlib import Path

from alfi.datasets import HomogeneousReactionDiffusion, ReactionDiffusionGenerator

"""
This script generates a synthetic reaction diffusion dataset .
"""


def save_dataset(toydata):
    """
    data_dir: the directory where the toy data and intermediate data lies. Will also be saved here.
    """
    data_dir = '../data'
    temp = pd.read_csv(Path(data_dir) / 'demToy1GPmRNA.csv').values
    t_sorted = np.argsort(temp[:, 0], kind='mergesort')
    # toydata = torch.load(Path(data_dir) / 'intermediate_toydata.pt')
    params_list = list()
    orig_data = list()
    num_samples = toydata[0]['samples'].shape[0]
    x_observed = torch.tensor(temp[t_sorted, 0:2]).permute(1, 0)

    for i in range(len(toydata)):
        params = torch.tensor([toydata[i][key] for key in ['l1', 'l2', 'sensitivity', 'decay', 'diffusion']])
        samples = toydata[i]['samples']
        for sample in range(num_samples):
            lf = samples[sample, 1681:]
            out = samples[sample, :1681]
            lf_out = torch.stack([lf, out], dim=0)
            orig_data.append(lf_out)
            params_list.append(params)
    params = torch.stack(params_list)
    orig_data = torch.stack(orig_data)
    shuffle = torch.randperm(orig_data.size()[0])
    orig_data = orig_data[shuffle]
    params = params[shuffle]
    torch.save({'x_observed': x_observed, 'orig_data': orig_data, 'params': params}, Path(data_dir) / 'toydata.pt')


if __name__ == '__main__':
    dataset = HomogeneousReactionDiffusion(data_dir='./data/', nn_format=False)
    tx = torch.tensor(dataset.orig_data[0]).t()
    with torch.no_grad():
        tot = 4 * 4 * 5 * 5
        i = 0
        objects = list()
        for sensitivity in np.linspace(0.2, 0.9, 6):
            for l1 in np.linspace(0.1, 0.4, 5):
                for l2 in np.linspace(0.1, 0.4, 5):
                    for diffusion in np.linspace(0.001, 0.1, 6):
                        for decay in np.linspace(0.01, 0.4, 5):
                            kernel = ReactionDiffusionGenerator(
                                lengthscale=[l1, l2],
                                decay=decay,
                                diffusion=diffusion,
                                sensitivity=sensitivity
                            )
                            Kuu, Kyy, Kyu, Kuy = kernel.joint(tx, tx)
                            kern = torch.zeros((2 * 1681, 2 * 1681))
                            kern[:1681, :1681] = Kyy
                            kern[:1681, 1681:] = Kyu
                            kern[1681:, :1681] = Kuy
                            kern[1681:, 1681:] = Kuu
                            try:
                                eigval, eigvec = torch.symeig(kern, eigenvectors=True)
                            except:
                                print('Failed for', l1, l2, sensitivity, decay, diffusion)
                                continue
                            eps = -1e-5
                            num = torch.sum((~(eigval >= eps)).type(torch.int)).item()
                            if num > 30:
                                print('Failed for', l1, l2, sensitivity, decay, diffusion)
                                continue
                            eigval_root = eigval.clamp_min(0.0).sqrt()
                            corr_matrix = (eigvec * eigval_root).transpose(-1, -2)
                            a = torch.randn(torch.Size([50, 2 * 1681]))
                            samples = a @ corr_matrix
                            obj = {
                                'samples': samples.clone(),
                                'sensitivity': sensitivity,
                                'l1': l1,
                                'l2': l2,
                                'diffusion': diffusion,
                                'decay': decay
                            }
                            objects.append(obj)
                            i += 1
                            print('Done ', i, '/', tot)

        # torch.save(, 'intermediate_toydata.pt')
        save_dataset(objects)
        print('Saved')



import torch
from torch.nn import Parameter
from matplotlib import pyplot as plt
from torch.optim import Adam
from gpytorch.optim import NGD
from gpytorch.constraints import Positive
import seaborn as sns
import numpy as np

from experiments.model_specs.variational import TranscriptionLFM
from alfi.datasets import P53Data
from alfi.configuration import VariationalConfiguration
from alfi.models import OrdinaryLFM, generate_multioutput_gp
from alfi.trainers import VariationalTrainer


""" 
Experiment for plotting the ideal number of inducing points 
"""


class P53ConstrainedTrainer(VariationalTrainer):
    def after_epoch(self):
        super().after_epoch()
        with torch.no_grad():
            self.lfm.basal_rate.clamp_(0, 20)
            self.lfm.decay_rate.clamp_(0, 20)
            sens = torch.tensor(1.)
            dec = torch.tensor(0.8)
            self.lfm.raw_sensitivity[3] = self.lfm.positivity.inverse_transform(sens)
            self.lfm.raw_decay[3] = self.lfm.positivity.inverse_transform(dec)


dataset = P53Data(replicate=0, data_dir='./data')
num_genes = 5
num_tfs = 1
config = VariationalConfiguration(
    preprocessing_variance=dataset.variance,
    num_samples=80,
    initial_conditions=False
)

def diff(lfm: TranscriptionLFM):
    B_exact, S_exact, D_exact = P53Data.params_ground_truth()
    B = lfm.basal_rate.detach().squeeze()
    D = lfm.decay_rate.detach().squeeze()
    S = lfm.sensitivity.detach().squeeze()
    mse = torch.square(B-B_exact) + torch.square(D-D_exact) + torch.square(S-S_exact)
    mse = mse.mean()

    return mse


with open('experiments/inducing_points.txt', 'w') as f:
    outputs = list()
    for i in range(5, 30):
        print('Running inducing points', i)
        num_inducing = i  # (I x m x 1)
        inducing_points = torch.linspace(0, 12, num_inducing).repeat(num_tfs, 1).view(num_tfs, num_inducing, 1)

        gp_model = generate_multioutput_gp(num_tfs, inducing_points, gp_kwargs=dict(natural=True))

        lfm = TranscriptionLFM(num_genes, gp_model, config)

        track_parameters = [
            'raw_basal',
            'raw_decay',
            'raw_sensitivity',
            'gp_model.covar_module.raw_lengthscale',
        ]
        num_training = dataset.m_observed.shape[-1]

        variational_optimizer = NGD(lfm.variational_parameters(), num_data=num_training, lr=0.1)
        parameter_optimizer = Adam(lfm.nonvariational_parameters(), lr=0.03)
        optimizers = [variational_optimizer, parameter_optimizer]
        trainer = P53ConstrainedTrainer(lfm, optimizers, dataset, track_parameters=track_parameters)

        lfm.train()
        trainer.train(350, report_interval=50, step_size=1e-1)
        last_loss = trainer.losses[-1].sum()
        mse = diff(lfm)
        f.write(f'{i}\t{mse}\n')
        outputs.append((i, mse, last_loss))

    outputs = np.array(outputs)
    palette = sns.color_palette('colorblind')
    line_color = palette[0]
    plt.style.use('seaborn')
    sns.set(font="CMU Serif")
    plt.rcParams['font.family'] = 'serif'
    plt.rcParams['font.serif'] = 'CMU Serif'

    plt.figure(figsize=(4, 3))
    plt.plot(outputs[:, 0], outputs[:, 2])
    plt.xlabel('Inducing points')
    plt.ylabel('MSE')
    plt.savefig('experiments/inducingpoints_loss.pdf')

    plt.figure(figsize=(4, 3))
    plt.plot(outputs[:, 0], outputs[:, 1])
    plt.xlabel('Inducing points')
    plt.ylabel('MSE')
    plt.savefig('experiments/inducingpoints_line.pdf')

    plt.figure(figsize=(4, 3))
    plt.scatter(outputs[:, 0], outputs[:, 1], s=5)
    plt.xlabel('Inducing points')
    plt.ylabel('MSE')
    plt.savefig('experiments/inducingpoints_scat.pdf')



from alfi.utilities.tf import get_time_square, FixedDistribution

import tensorflow as tf
from tensorflow import math as tfm
from tensorflow_probability import distributions as tfd
import numpy as np

f64 = np.float64


class GPKernelSelector():
    def __init__(self, data, options):
        self.kernel = options.kernel
        self.options = options
        self.t_discretised = data.t_discretised
        self.N_p = data.t_discretised.shape[0]
        self.num_tfs = data.f_obs.shape[1]
        t_1, t_2 = get_time_square(self.t_discretised, self.N_p)
        self.t_dist = t_1-t_2
        self.tt = t_1*t_2
        self.t2 = tf.square(t_1)
        self.tprime2 = tf.square(t_2)
        self.fixed_dist = FixedDistribution(tf.ones(self.num_tfs, dtype='float64'))
        min_dist = min(data.t_observed[1:] - data.t_observed[:-1])
        min_dist = max(min_dist, 1.)
        self._ranges = {
            'rbf': [(f64(1e-4), f64(5)),  #1+max(np.var(data.f_obs, axis=2))
                    (f64(min_dist**2)-1.2, f64(data.t_observed[-1] ** 2))],
            'mlp': [(f64(1), f64(10)), (f64(3.5), f64(20))],
        }
        self._priors = {
            'rbf': [tfd.Uniform(f64(1), f64(20)), tfd.Uniform(f64(min_dist**2), f64(10))],
            'mlp': [tfd.Uniform(f64(3.5), f64(10)), tfd.InverseGamma(f64(0.01), f64(0.01))],
        }
        v_prop = lambda v: tfd.TruncatedNormal(v, 0.01, low=0, high=100)
        l2_prop = lambda l2: tfd.TruncatedNormal(l2, 0.01, low=0, high=100)
        proposals = [v_prop, l2_prop]
        self._proposals = {
            'rbf': proposals,
        }
        self._names = {'rbf': ['v','l2'], 'mlp': ['w', 'b']}

    def __call__(self):
        '''Calculates kernel covariance matrix'''
        if self.kernel == 'rbf':
            return self.rbf
        elif self.kernel == 'mlp':
            return self.mlp
        else:
            raise Exception('No kernel by that name!')

    def initial_params(self):
        if self.kernel == 'rbf':
            return [1.1*tf.ones(self.num_tfs, dtype='float64'), 
                    2*tf.ones(self.num_tfs, dtype='float64')]
        elif self.kernel == 'mlp':
            return [0.8*tf.ones(self.num_tfs, dtype='float64'), 
                    0.98*tf.ones(self.num_tfs, dtype='float64')]

    def ranges(self):
        return self._ranges[self.kernel]
    def names(self):
        return self._names[self.kernel]
    def priors(self):
        return self._priors[self.kernel]

    def proposal(self, hyp_index, current_val):
        '''Returns kernel hyperparameter proposal dist centred on current val'''
        if hyp_index == 0 and not self.options.translation:
            return self.fixed_dist
        return self._proposals[self.kernel][hyp_index](current_val)

    def rbf(self, v, l2):
        if self.options.kernel_exponential:
            v = tf.exp(v)
            l2 = tf.exp(l2)
        sq_dist = tf.divide(tfm.square(self.t_dist), tf.reshape(2*l2, (-1, 1, 1)))
        K = tf.reshape(v, (-1, 1, 1)) * tfm.exp(-sq_dist)
        m = tf.zeros((self.N_p), dtype='float64')
        return m, K

    def mlp(self, w, b):
        w = tf.reshape(w, (-1, 1, 1))
        b = tf.reshape(b, (-1, 1, 1))
        denom = tfm.sqrt((w*self.t2 + b + 1) * (w*self.tprime2 + b + 1))
        K = tfm.asin((w*self.tt + b)/denom)
        m = tf.zeros((self.N_p), dtype='float64')
        return m, K




import torch
import numpy as np
import pandas as pd

from pathlib import Path
from tqdm import tqdm
from os import path
from abc import ABC
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from scipy.integrate import odeint
from scipy.io import loadmat

from alfi.datasets import load_barenco_puma
from alfi.utilities.data import generate_neural_dataset_2d
from . import LFMDataset


f64 = np.float64


class TranscriptomicTimeSeries(LFMDataset, ABC):
    def __init__(self):
        self._m_observed = None
        self._t_observed = None

    @property
    def t_observed(self):
        return self._t_observed

    @t_observed.setter
    def t_observed(self, value):
        self._t_observed = value

    @property
    def m_observed(self):
        """m_observed has shape (replicates, genes, times)"""
        return self._m_observed

    @m_observed.setter
    def m_observed(self, value):
        self._m_observed = value


class P53Data(TranscriptomicTimeSeries):
    def __init__(self, replicate=None, data_dir='../data/'):
        super().__init__()
        m_observed, f_observed, σ2_m_pre, σ2_f_pre, t = load_barenco_puma(data_dir)

        m_df, m_observed = m_observed  # (replicates, genes, times)
        self.gene_names = m_df.index
        num_times = m_observed.shape[2]
        num_genes = m_observed.shape[1]
        num_replicates = m_observed.shape[0]
        self.num_outputs = num_genes

        # f_df, f_observed = f_observed
        m_observed = torch.tensor(m_observed)
        self.t_observed = torch.linspace(f64(0), f64(12), 7)
        self.m_observed = m_observed
        self.f_observed = torch.tensor([0.7116,0.7008,1.5933,0.7507,0.2346,0.3617,0.0673]).view(1, 1, 7)
        self.f_observed = torch.tensor([0.1845,1.1785,1.6160,0.8156,0.6862,-0.1828, 0.5131]).view(1, 1, 7)
        self.ups_f_observed = torch.tensor([ 0.1654,  0.2417,  0.3680,  0.5769,  0.8910,  1.2961,  1.7179,  2.0301,
         2.1055,  1.8885,  1.4407,  0.9265,  0.5467,  0.4493,  0.6573,  1.0511,
         1.4218,  1.5684,  1.3876,  0.9156,  0.3090, -0.2256, -0.5218, -0.5191,
        -0.2726,  0.0860,  0.4121,  0.6056,  0.6397,  0.5517])
        self.ups_m_observed = torch.tensor([
            [ 0.3320,  0.0215,  0.0643,  0.0200,  0.2439],
            [ 0.3930,  0.0870,  0.1319,  0.0946,  0.3119],
            [ 0.4726,  0.1710,  0.2184,  0.1799,  0.3989],
            [ 0.5886,  0.2926,  0.3435,  0.2975,  0.5247],
            [ 0.7662,  0.4779,  0.5343,  0.4745,  0.7165],
            [ 1.0297,  0.7522,  0.8165,  0.7324,  1.0003],
            [ 1.3845,  1.1195,  1.1942,  1.0670,  1.3800],
            [ 1.8027,  1.5484,  1.6352,  1.4351,  1.8230],
            [ 2.2269,  1.9766,  2.0751,  1.7660,  2.2647],
            [ 2.5875,  2.3304,  2.4380,  1.9863,  2.6287],
            [ 2.8234,  2.5466,  2.6588,  2.0439,  2.8495],
            [ 2.8991,  2.5906,  2.7022,  1.9238,  2.8917],
            [ 2.8242,  2.4765,  2.5830,  1.6655,  2.7703],
            [ 2.6636,  2.2771,  2.3761,  1.3663,  2.5610],
            [ 2.5173,  2.1008,  2.1935,  1.1477,  2.3765],
            [ 2.4731,  2.0404,  2.1307,  1.0952,  2.3129],
            [ 2.5574,  2.1206,  2.2130,  1.2074,  2.3954],
            [ 2.7229,  2.2873,  2.3844,  1.3978,  2.5677],
            [ 2.8762,  2.4384,  2.5395,  1.5417,  2.7234],
            [ 2.9200,  2.4701,  2.5715,  1.5321,  2.7550],
            [ 2.7960,  2.3228,  2.4191,  1.3264,  2.6011],
            [ 2.5121,  2.0090,  2.0952,  0.9666,  2.2748],
            [ 2.1368,  1.6051,  1.6790,  0.5574,  1.8559],
            [ 1.7647,  1.2137,  1.2763,  0.2150,  1.4509],
            [ 1.4744,  0.9177,  0.9722,  0.0158,  1.1455],
            [ 1.3047,  0.7557,  0.8064, -0.0219,  0.9796],
            [ 1.2526,  0.7211,  0.7719,  0.0694,  0.9458],
            [ 1.2842,  0.7743,  0.8278,  0.2286,  1.0029],
            [ 1.3540,  0.8650,  0.9222,  0.3933,  1.0986],
            [ 1.4224,  0.9504,  1.0107,  0.5194,  1.1882]]).t()

        self.ups_t_observed = t_predict = torch.linspace(0, 12, 30, dtype=torch.float32)

        self.params = torch.tensor([
            0.06374478, 0.01870999, 0.0182909,  0.0223461,  0.08485352, 0.9133557, 0.9743523,
            0.9850107,  1., 0.974792,   0.27596828, 0.367931, 0.35159853, 0.79999995, 0.34772962
        ]).view(3, 5)
        if replicate is None:
            self.variance = np.array([f64(σ2_m_pre)[r, i] for r in range(num_replicates) for i in range(num_genes)])
            self.data = [(self.t_observed, m_observed[r, i]) for r in range(num_replicates) for i in range(num_genes)]
        else:
            self.m_observed = self.m_observed[replicate:replicate+1]
            self.f_observed = self.f_observed[0:1]
            self.variance = np.array([f64(σ2_m_pre)[replicate, i] for i in range(num_genes)])
            self.data = [(self.t_observed, m_observed[replicate, i]) for i in range(num_genes)]

    @staticmethod
    def params_ground_truth():
        B_exact = np.array([0.0649, 0.0069, 0.0181, 0.0033, 0.0869])
        D_exact = np.array([0.2829, 0.3720, 0.3617, 0.8000, 0.3573])
        S_exact = np.array([0.9075, 0.9748, 0.9785, 1.0000, 0.9680])
        return B_exact, S_exact, D_exact


class HafnerData(TranscriptomicTimeSeries):
    """
    Dataset of GSE100099
    MCF7 cells gamma-irradiated over 24 hours
    p53 is typically the protein of interest
    """
    def __init__(self, replicate=None, data_dir='../data/', extra_targets=True):
        super().__init__()
        target_genes = [
            'KAZN','PMAIP1','PRKAB1','CSNK1G1','E2F7','SLC30A1',
            'PTP4A1','RAP2B','SUSD6','UBR5-AS1','RNF19B','AEN','ZNF79','XPC',
            'FAM212B','SESN2','DCP1B','MDM2','GADD45A','SESN1','CDKN1A','BTG2'
        ]
        if extra_targets:
            target_genes.extend([
                'DSCAM','C14orf93','RPL23AP64','RPS6KA5','MXD1', 'LINC01560', 'THNSL2',
                'EPAS1', 'ARSD', 'NACC2', 'NEDD9', 'GATS', 'ABHD4', 'BBS1', 'TXNIP',
                'KDM4A', 'ZNF767P', 'LTB4R', 'PI4K2A', 'ZNF337', 'PRKX', 'MLLT11',
                'HSPA4L', 'CROT', 'BAX', 'ORAI3', 'CES2', 'PVT1', 'ZFYVE1', 'PIK3R3',
                'TSPYL2', 'PROM2', 'ZBED5-AS1', 'CCNG1', 'STOM','IER5','STEAP3',
                'TYMSOS','TMEM198B','TIGAR','ASTN2','ANKRA2','RRM2B','TAP1','TP53I3','PNRC1',
                'GLS2','TMEM229B','IKBIP','ERCC5','KIAA1217','DDIT4','DDB2','TP53INP1'
            ])
        # np.random.shuffle(target_genes)
        self.num_outputs = len(target_genes)
        tfs = ['TP53']
        with open(Path(data_dir) / 'GSE100099_RNASeqGEO.tsv', 'r', 1) as f:
            contents = f.buffer
            df = pd.read_table(contents, sep='\t', index_col=0)

        columns = ['MCF7, t=0 h, rep1']
        columns.extend(['MCF7, t='+str(t)+' h, IR 10Gy, rep1' for t in range(1, 13)])
        columns.append('MCF7, t=0 h, rep1')
        columns.extend(['MCF7, t=' + str(t) + ' h, IR 10Gy, rep2' for t in range(1, 13)])
        # This dataset only has two complete replicates

        self.genes_df = df[df.index.isin(target_genes)][columns]
        self.genes_df = self.genes_df.reindex(target_genes)
        self.tfs_df = df[df.index.isin(tfs)][columns]

        m = self.genes_df.values
        genes_norm = 1/m.shape[0] * np.linalg.norm(m, axis=1, ord=None)  # l2 norm
        self.m_observed = torch.tensor(m / np.sqrt(genes_norm.reshape(-1, 1)), dtype=torch.float32)

        f = self.tfs_df.values
        tfs_norm = 1/f.shape[0] * np.linalg.norm(f, axis=1, ord=None)  # l2 norm
        self.tfs = (f / np.sqrt(tfs_norm.reshape(-1, 1)))
        self.tfs = self.tfs.reshape((1, 2, 13)).swapaxes(0,1)

        self.t_observed = torch.linspace(0, 12, 13, dtype=torch.float32)
        self.m_observed = self.m_observed.reshape(self.num_outputs, 2, 13).transpose(0, 1)

        if replicate is None:
            self.data = [(self.t_observed, self.m_observed[r, i]) for r in range(2) for i in range(self.num_outputs)]
        else:
            self.data = [(self.t_observed, self.m_observed[replicate, i]) for i in range(self.num_outputs)]

        self.gene_names = np.array(target_genes)


class MCMCToyTimeSeries(TranscriptomicTimeSeries):
    def __init__(self, delay=False):
        # We import the dataset here since it uses TensorFlow
        from alfi.datasets.artificial import get_artificial_dataset
        super().__init__()
        nodelay_dataset, delay_dataset = get_artificial_dataset()
        p_nodelay, m_nodelay = nodelay_dataset
        replicate = 0
        m_nodelay = m_nodelay[replicate]
        p_nodelay = p_nodelay[replicate]
        self.num_genes = m_nodelay.shape[0]
        self.num_tfs = p_nodelay.shape[0]
        self.f_observed = p_nodelay
        num_times = m_nodelay.shape[1]

        self.gene_names = np.arange(self.num_genes)
        self.m_observed = torch.tensor(m_nodelay).unsqueeze(0)
        self.t = torch.linspace(f64(0), f64(1), num_times, dtype=torch.float64)
        self.data = [(self.t, self.m_observed[0, i]) for i in range(self.num_genes)]

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return 1


class DrosophilaSpatialTranscriptomics(LFMDataset):
    """
    Dataset from Becker et al. (2013).
    Reverse engineering post-transcriptional regulation of
    gap genes in Drosophila melanogaster
    """
    def __init__(self, gene='kr', data_dir='../data/', scale=False, scale_tx=False, nn_format=False, disc=1):
        indents = {'kr': 64, 'kni': 56, 'gt': 60}
        assert gene in indents
        data = pd.read_csv(path.join(data_dir, f'clean_{gene}.csv'))
        data = data.iloc[indents[gene]:].values
        data = data[:, [0, 1, 3, 2]]
        if scale:
            scaler = StandardScaler()
            data[:, 2:3] = scaler.fit_transform(data[:, 2:3])
            data[:, 3:4] = scaler.transform(data[:, 3:4])
        if scale_tx:
            scaler = MinMaxScaler()
            data[:, 0:1] = scaler.fit_transform(data[:, 0:1])
            data[:, 1:2] = scaler.fit_transform(data[:, 1:2])

        self.orig_data = torch.tensor(data).t()
        self.num_outputs = 1
        self.disc = disc
        self.num_discretised = self.disc*7
        x_observed = torch.tensor(data[:, 0:2]).permute(1, 0)
        data = torch.tensor(data[:, 3]).unsqueeze(0)
        self.gene_names = np.array([gene])
        if nn_format:
            params = torch.tensor([-1.]*5).unsqueeze(0)
            train, test = generate_neural_dataset_2d(self.orig_data.unsqueeze(0), params, 1, 0)
            self.data = train
            self.train_data = train
            self.test_data = test
        else:
            self.data = [(x_observed, data)]


class MarkovJumpProcess:
    """
    Implements a generic markov jump process and algorithms for simulating it.
    It is an abstract class, it needs to be inherited by a concrete implementation.
    """

    def __init__(self, init, params):

        self.state = np.asarray(init)
        self.params = np.asarray(params)
        self.time = 0.0

    def _calc_propensities(self):
        raise NotImplementedError('This is an abstract method and should be implemented in a subclass.')

    def _do_reaction(self, reaction):
        raise NotImplementedError('This is an abstract method and should be implemented in a subclass.')

    def sim_steps(self, num_steps):
        """Simulates the process with the gillespie algorithm for a specified number of steps."""

        times = [self.time]
        states = [self.state.copy()]

        for _ in range(num_steps):

            rates = self.params * self._calc_propensities()
            total_rate = rates.sum()

            if total_rate == 0:
                self.time = float('inf')
                break

            self.time += np.random.exponential(scale=1 / total_rate)

            reaction = self.discrete_sample(rates / total_rate)[0]
            self._do_reaction(reaction)

            times.append(self.time)
            states.append(self.state.copy())

        return times, np.array(states)

    def sim_time(self, dt, duration, max_n_steps=float('inf')):
        """Simulates the process with the gillespie algorithm for a specified time duration."""

        num_rec = int(duration / dt) + 1
        states = np.zeros([num_rec, self.state.size])
        cur_time = self.time
        n_steps = 0

        for i in range(num_rec):

            while cur_time > self.time:

                rates = self.params * self._calc_propensities()
                total_rate = rates.sum()

                if total_rate == 0:
                    self.time = float('inf')
                    break

                exp_scale = max(1 / total_rate, 1e-3)
                self.time += np.random.exponential(scale=exp_scale)

                reaction = np.random.multinomial(1, rates / total_rate)
                reaction = np.argmax(reaction)
                self._do_reaction(reaction)

                n_steps += 1
                if n_steps > max_n_steps:
                    raise SimTooLongException(max_n_steps)

            states[i] = self.state.copy()
            cur_time += dt

        return np.array(states)


class LotkaVolterra(MarkovJumpProcess):
    """Implements the lotka-volterra population model."""

    def _calc_propensities(self):

        x, y = self.state
        xy = x * y
        return np.array([xy, x, y, xy])

    def _do_reaction(self, reaction):

        if reaction == 0:
            self.state[0] += 1
        elif reaction == 1:
            self.state[0] -= 1
        elif reaction == 2:
            self.state[1] += 1
        elif reaction == 3:
            self.state[1] -= 1
        else:
            raise ValueError('Unknown reaction.')


class StochasticLotkaVolteraData(LFMDataset):
    """
    Dataset of time-seires sampled from a Lotka-Voltera model
    ----------
    amplitude_range : tuple of float
        Defines the range from which the amplitude (i.e. a) of the sine function
        is sampled.
    shift_range : tuple of float
        Defines the range from which the shift (i.e. b) of the sine function is
        sampled.
    num_samples : int
        Number of samples of the function contained in dataset.
    num_points : int
        Number of points at which to evaluate f(x) for x in [-pi, pi].
    """

    def __init__(self, initial_X=50, initial_Y=100,
                 num_samples=1000, dt=0.2):
        self.initial_X = initial_X
        self.initial_Y = initial_Y
        self.num_samples = num_samples
        self.x_dim = 1
        self.y_dim = 2
        self.dt = dt

        self.init = [self.initial_X, self.initial_Y]
        self.params = [0.01, 0.5, 1.0, 0.01]
        self.duration = 30

        # Generate data
        self.data = []
        print("Creating dataset...", flush=True)

        removed = 0
        for samples in range(num_samples):
            lv = LotkaVolterra(self.init, self.params)
            states = lv.sim_time(dt, self.duration)
            times = torch.linspace(0.0, self.duration,
                                   int(self.duration / dt) + 1)
            times = times.unsqueeze(1)

            # Ignore outlier populations
            if np.max(states) > 600:
                removed += 1
                continue

            # Scale the population ranges to be closer to the real model
            states = torch.FloatTensor(states) * 1 / 100
            times = times * 1 / 20
            self.data.append((times, states))

        self.num_samples -= removed

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return self.num_samples


class DeterministicLotkaVolteraData(LFMDataset):
    """
    Dataset of Lotka-Voltera time series.
      Populations (u,v) evolve according to
        u' = \alpha u - \beta u v
        v' = \delta uv - \gamma v
      with the dataset sampled either with (u_0, v_0) fixed and (\alpha, \beta,
      \gamma, \delta) varied, or varying the initial populations for a fixed set
      of greeks.
    If initial values for (u,v) are provided then the greeks are sampled from
        (0.9,0.05,1.25,0.5) to (1.1,0.15,1.75,1.0)
    if values are provided for the greeks then (u_0 = v_0) is sampled from
        (0.5) to (2.0)
    if both are provided, defaults to initial population mode (greeks vary)
    ----------
    initial_u	: int
        fixed initial value for u
    initial_v	: int
        fixed initial value for v
    fixed_alpha : int
        fixed initial value for \alpha
    fixed_beta	: int
        fixed initial value for \beta
    fixed_gamma : int
        fixed initial value for \gamme
    fixed_delta : int
        fixed initial value for \delta
    end_time : float
        the final time (simulation runs from 0 to end_time)
    steps : int
        how many time steps to take from 0 to end_time
    num_samples : int
        Number of samples of the function contained in dataset.
    """

    def __init__(self, initial_u=None, initial_v=None,
                 alpha=None, beta=None, gamma=None, delta=None,
                 num_samples=1000, steps=150, end_time=15):

        if initial_u is None:
            self.mode = 'greek'
            self.alpha = alpha
            self.beta = beta
            self.gamma = gamma
            self.delta = delta
        else:
            self.mode = 'population'
            self.initial_u = initial_u
            self.initial_v = initial_v

        print('Lotka-Voltera is in {self.mode} mode.')

        self.num_samples = num_samples
        self.steps = steps
        self.end_time = end_time

        # Generate data
        self.data = []
        print("Creating dataset...", flush=True)

        removed = 0
        for samples in tqdm(range(num_samples)):
            times, states = self.generate_ts()
            #  normalise times
            times = torch.FloatTensor(times) / 10
            times = times.unsqueeze(1)

            states = torch.FloatTensor(states)
            if self.mode == 'population':
                states = states / 100
            # states = torch.cat((states, times), dim=-1)

            self.data.append((times, states))

        self.num_samples -= removed

    def generate_ts(self):
        if self.mode == 'population':
            X_0 = np.array([self.initial_u, self.initial_v])
            a = np.random.uniform(0.9, 1.1)
            b = np.random.uniform(0.05, 0.15)
            c = np.random.uniform(1.25, 1.75)
            d = np.random.uniform(0.5, 1.0)
        else:
            equal_pop = np.random.uniform(0.25, 1.)
            X_0 = np.array([2 * equal_pop, equal_pop])
            a, b, c, d = self.alpha, self.beta, self.gamma, self.delta

        def dX_dt(X, t=0):
            """ Return the growth rate of fox and rabbit populations. """
            return np.array([a * X[0] - b * X[0] * X[1],
                             -c * X[1] + d * X[0] * X[1]])

        t = np.linspace(0, self.end_time, self.steps)
        X = odeint(dX_dt, X_0, t)

        return t, X

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return self.num_samples


class RotNISTDataset(LFMDataset):
    """
    Loads the rotated 3s from ODE2VAE paper
    https://www.dropbox.com/s/aw0rgwb3iwdd1zm/rot-mnist-3s.mat?dl=0
    """
    def __init__(self, data_dir='../data'):
        mat = loadmat(path.join(data_dir, 'rot-mnist-3s.mat'))
        dataset = mat['X'][0]
        dataset = dataset.reshape(dataset.shape[0], dataset.shape[1], -1)
        self.data = torch.tensor(dataset, dtype=torch.float32)
        self.t = torch.linspace(0, 1, dataset.shape[1], dtype=torch.float32).view(-1, 1).repeat([dataset.shape[0], 1, 1])
        self.data = list(zip(self.t, self.data))

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return len(self.data)



from abc import ABC
from enum import Enum

import torch

from gpytorch.models import ApproximateGP
from gpytorch.likelihoods import MultitaskGaussianLikelihood

from .lfm import LFM
from alfi.configuration import VariationalConfiguration
from alfi.mlls import MaskedVariationalELBO
from alfi.utilities import is_cuda


class TrainMode(Enum):
    NORMAL = 0
    """
    Gradient matching: the derivative of the spline interpolation is matched with the output of the ODE function
    """
    GRADIENT_MATCH = 1
    """
    Filter: the output of the ODE function is matched with the filtered data bucketed and filtered into the same
    timepoints
    """
    FILTER = 2
    """
    Samples: return samples from the ODE
    """
    SAMPLES = 3


class VariationalLFM(LFM, ABC):
    """
    Variational inducing point approximation for Latent Force Models.

    Parameters
    ----------
    num_outputs : int : the number of outputs (for example, the number of genes)
    fixed_variance : tensor : variance if the preprocessing variance is known, otherwise learnt.
    """
    def __init__(self,
                 num_outputs: int,
                 gp_model: ApproximateGP,
                 config: VariationalConfiguration,
                 num_training_points=None,
                 dtype=torch.float64):
        super().__init__()
        self.gp_model = gp_model
        self.num_outputs = num_outputs
        self.train_mode = TrainMode.NORMAL
        self.config = config
        self.dtype = dtype

        try:
            self.inducing_points = self.gp_model.get_inducing_points()
        except AttributeError:
            raise AttributeError('The GP model must define a function `get_inducing_points`.')

        # Construct likelihood
        self.num_tasks = num_outputs
        if num_training_points is None:
            num_training_points = self.inducing_points.numel()  # TODO num_data refers to the number of training datapoints

        self.num_latents = gp_model.variational_strategy.num_tasks
        if config.latent_data_present:  # add latent force likelihood
            self.num_tasks += self.num_latents

        self.likelihood = MultitaskGaussianLikelihood(num_tasks=self.num_tasks)

        self.loss_fn = MaskedVariationalELBO(self.likelihood, gp_model, num_training_points, combine_terms=False)

    def nonvariational_parameters(self):
        variational_keys = dict(self.gp_model.named_variational_parameters()).keys()
        named_parameters = dict(self.named_parameters())
        return [named_parameters[key] for key in named_parameters.keys()
                if key[len('gp_model.'):] not in variational_keys]

    def variational_parameters(self):
        return self.gp_model.variational_parameters()

    def summarise_gp_hyp(self):
        with torch.no_grad():
            def convert(x):
                inducing_points = self.inducing_points.cuda() if is_cuda() else self.inducing_points
                x = x.detach().cpu().view(-1).numpy()[:5]
                # noise = self.gp_model(inducing_points).variance.mean(dim=0).cpu()
                return str(x)  #+ ' noise: ' + str(noise)
            if self.gp_model.covar_module.lengthscale is not None:
                return convert(self.gp_model.covar_module.lengthscale)
            elif hasattr(self.gp_model.covar_module, 'base_kernel'):
                kernel = self.gp_model.covar_module.base_kernel
                if hasattr(kernel, 'kernels'):
                    if hasattr(kernel.kernels[0], 'lengthscale'):
                        return convert(kernel.kernels[0].lengthscale)
                else:
                    return convert(self.gp_model.covar_module.base_kernel.lengthscale)
            else:
                return ''

    def forward(self, x, **kwargs):
        raise NotImplementedError

    def train(self, mode: bool = True):
        self.gp_model.train(mode)
        self.likelihood.train(mode)

    def set_mode(self, mode=TrainMode.NORMAL):
        self.train_mode = mode

    def eval(self):
        self.gp_model.eval()
        self.likelihood.eval()
        if self.train_mode == TrainMode.GRADIENT_MATCH or self.train_mode == TrainMode.FILTER:
            self.set_mode(TrainMode.NORMAL)

    def predict_m(self, t_predict, **kwargs) -> torch.distributions.MultivariateNormal:
        """
        Calls self on input `t_predict`
        """
        return self(t_predict.view(-1), **kwargs)

    def predict_f(self, t_predict, **kwargs) -> torch.distributions.MultivariateNormal:
        """
        Returns the latents
        """
        self.eval()
        with torch.no_grad():
            q_f = self.gp_model(t_predict)
        self.train()
        return q_f

    def save(self, filepath):
        torch.save(self.gp_model.state_dict(), str(filepath)+'gp.pt')
        torch.save(self.state_dict(), str(filepath)+'lfm.pt')

    @classmethod
    def load(cls,
             filepath,
             gp_cls=None, gp_model=None,
             gp_args=[], gp_kwargs={},
             lfm_args=[], lfm_kwargs={}):
        assert not (gp_cls is None and (gp_model is None))
        gp_state_dict = torch.load(filepath+'gp.pt')
        if gp_cls is not None:
            gp_model = gp_cls(*gp_args, **gp_kwargs)
        gp_model.load_state_dict(gp_state_dict)
        gp_model.double()

        lfm_state_dict = torch.load(filepath+'lfm.pt')
        lfm = cls(lfm_args[0], gp_model, *lfm_args[1:], **lfm_kwargs)
        lfm.load_state_dict(lfm_state_dict)
        return lfm
